#!/usr/bin/env python3
"""Design guides for diagnostics."""

import argparse
from collections import defaultdict
import logging
import os
import random
import re
import shutil
import sys
import tempfile

import numpy as np

from adapt import alignment
from adapt import guide_search
from adapt.prepare import align
from adapt.prepare import ncbi_neighbors
from adapt.prepare import prepare_alignment
from adapt import primer_search
from adapt.specificity import alignment_query
from adapt import target_search
from adapt.utils import log
from adapt.utils import predict_activity
from adapt.utils import seq_io
from adapt.utils import year_cover
from adapt.utils import mutate
from adapt.utils.version import get_project_path, get_latest_model_version

try:
    import boto3
    from botocore.exceptions import ClientError
except ImportError:
    cloud = False
else:
    cloud = True

__author__ = 'Hayden Metsky <hmetsky@broadinstitute.org>, Priya P. Pillai <ppillai@broadinstitute.org>'

logger = logging.getLogger(__name__)


# Define defaults for parameters associated with different objectives
# The reason to define them here, rather than in argparse, is that
# defining them outside of argparse lets us check whether the
# argument has been set (versus just set to its default)
OBJ_PARAM_DEFAULTS = {
        'minimize-guides': {
            'guide_mismatches': 0,
            'guide_cover_frac': 1.0
        },
        'maximize-activity': {
            'guide_mismatches': 0,
            'soft_guide_constraint': 1,
            'hard_guide_constraint': 5,
            'penalty_strength': 0.25,
            'maximization_algorithm': 'random-greedy'
        }
}


def check_obj_args(args):
    """Check argument depending on the objective function and set defaults.

    Raise critical messages if arguments are set for the wrong objective
    function than what is specified, and set defaults for these arguments if
    they are not specified

    Args:
        args: namespace of arguments provided to this executable
    """
    if args.obj == 'minimize-guides':
        if (args.soft_guide_constraint or args.hard_guide_constraint or
                args.penalty_strength or args.maximization_algorithm):
            logger.critical(("When --obj is 'minimize-guides', the "
                "following arguments are not used: --soft-guide-constraint, "
                "--hard-guide-constraint, --penalty-strength, "
                "--maximization-algorithm"))
        if args.use_simple_binary_activity_prediction:
            raise Exception(("Cannot use --use-simple-binary-activity-prediction "
                "when --obj is 'minimize-guides'"))
        for arg in OBJ_PARAM_DEFAULTS['minimize-guides'].keys():
            if vars(args)[arg] is None:
                vars(args)[arg] = OBJ_PARAM_DEFAULTS['minimize-guides'][arg]

    if args.obj == 'maximize-activity':
        if args.use_simple_binary_activity_prediction:
            if (args.guide_cover_frac or args.cover_by_year_decay):
                logger.critical(("When --obj is 'maximize-activity', the "
                    "following arguments are not used: "
                    "--guide-cover-frac, --cover-by-year-decay"))
            if args.predict_activity_model_path:
                logger.critical(("When --use-simple-binary-activity-prediction "
                    "is set, --predict-activity-model-path is not used"))
            if args.predict_cas13a_activity_model:
                logger.critical(("When --use-simple-binary-activity-prediction "
                    "is set, --predict-cas13a-activity-model is not used"))
        else:
            if (args.guide_mismatches or args.guide_cover_frac or
                    args.cover_by_year_decay):
                logger.critical(("When --obj is 'maximize-activity', the "
                    "following arguments are not used: --guide-mismatches, "
                    "--guide-cover-frac, --cover-by-year-decay"))
        for arg in OBJ_PARAM_DEFAULTS['maximize-activity'].keys():
            if vars(args)[arg] is None:
                vars(args)[arg] = OBJ_PARAM_DEFAULTS['maximize-activity'][arg]


def seqs_grouped_by_year(seqs, args):
    """Group sequences according to their year and assigned partial covers.

    Args:
        seqs: dict mapping sequence name to sequence, as read from a FASTA
        args: namespace of arguments provided to this executable

    Returns:
        tuple (aln, years_idx, cover_frac) where aln is an
        alignment.Alignment object from seqs; years_idx is a dict
        mapping each year to the set of indices in aln representing
        sequences for that year; and cover_frac is a dict mapping each
        year to the desired partial cover of sequences from that year,
        as determined by args.cover_by_year_decay (if args.search_cmd
        is 'complete-targets', then cover_frac is a tuple (g, p) where
        g is the previously described dict calculated from
        args.guide_cover_frac and p is the same calculated from
        args.primer_cover_frac)
    """
    years_fn, year_highest_cover, year_cover_decay = args.cover_by_year_decay

    # Map sequence names to index in alignment, and construct alignment
    seq_list = []
    seq_idx = {}
    for i, (name, seq) in enumerate(seqs.items()):
        seq_idx[name] = i
        seq_list += [seq]
    aln = alignment.Alignment.from_list_of_seqs(seq_list)

    # Read sequences for each year, and check that every sequence has
    # a year
    years = year_cover.read_years(years_fn)
    all_seqs_with_year = set.union(*years.values())
    for seq in seq_idx.keys():
        if seq not in all_seqs_with_year:
            raise Exception("Unknown year for sequence '%s'" % seq)

    # Convert years dict to map to indices rather than sequence names
    years_idx = {}
    for year in years.keys():
        # Skip names not in seq_idx because the years file may contain
        # sequences that are not in seqs
        years_idx[year] = set(seq_idx[name] for name in years[year]
            if name in seq_idx)

    # Construct desired partial cover for each year
    guide_cover_frac = year_cover.construct_partial_covers(
        years.keys(), year_highest_cover, args.guide_cover_frac, year_cover_decay)

    if args.search_cmd == 'complete-targets':
        primer_cover_frac = year_cover.construct_partial_covers(
            years.keys(), year_highest_cover, args.primer_cover_frac, year_cover_decay)
        cover_frac = (guide_cover_frac, primer_cover_frac)
    else:
        cover_frac = guide_cover_frac

    return aln, years_idx, cover_frac


def parse_required_guides_and_blacklist(args):
    """Parse files giving required guides and blacklisted sequence.

    Args:
        args: namespace of arguments provided to this executable

    Returns:
        tuple (required_guides, blacklisted_ranges) where required_guides
        is a representation of data in the args.required_guides file;
        blacklisted_ranges is a representation of data in the
        args.blacklisted_ranges file; and blacklisted_kmers is a
        representation of data in the args.blacklisted_kmers file
    """
    num_aln = len(args.in_fasta)

    # Read required guides, if provided
    if args.required_guides:
        required_guides = seq_io.read_required_guides(
            args.required_guides, args.guide_length, num_aln)
    else:
        required_guides = [{} for _ in range(num_aln)]

    # Read blacklisted ranges, if provided
    if args.blacklisted_ranges:
        blacklisted_ranges = seq_io.read_blacklisted_ranges(
            args.blacklisted_ranges, num_aln)
    else:
        blacklisted_ranges = [set() for _ in range(num_aln)]

    # Read blacklisted kmers, if provided
    if args.blacklisted_kmers:
        blacklisted_kmers = seq_io.read_blacklisted_kmers(
            args.blacklisted_kmers,
            min_len_warning=5,
            max_len_warning=args.guide_length)
    else:
        blacklisted_kmers = set()

    return required_guides, blacklisted_ranges, blacklisted_kmers


def prepare_alignments(args):
    """Download, curate, and align sequences for input.

    Args:
        args: namespace of arguments provided to this executable

    Returns:
        tuple (in_fasta, taxid_for_fasta, years_tsv, aln_tmp_dirs, out_tsv, design_for)
        in which in_fasta is a list of paths to fasta files each containing an
        alignment, taxid_for_fasta[i] gives a taxon id for in_fasta[i],
        years_tsv gives a tempfile storing a tsv file containing a year for
        each sequence across all the fasta files (only if
        args.cover_by_year_decay is set), aln_tmp_dirs is a list of temp
        directories that need to be cleaned up, out_tsv[i] is a path to the
        file at which to write the output for in_fasta[i], and design_for[i]
        indicates whether to actually design for in_fasta[i] (True/False)
    """
    logger.info(("Setting up to prepare alignments"))

    # Set the path to mafft
    align.set_mafft_exec(args.mafft_path)

    # Setup alignment and alignment stat memoizers
    if args.prep_memoize_dir:
        if args.prep_memoize_dir[:5] == "s3://":
            bucket = args.prep_memoize_dir.split("/")[2]
            try:
                if args.aws_access_key_id is not None and args.aws_secret_access_key is not None:
                    S3 = boto3.client("s3", aws_access_key_id=args.aws_access_key_id,
                        aws_secret_access_key=args.aws_secret_access_key)
                else:
                    S3 = boto3.client("s3")
                S3.head_bucket(Bucket=bucket)
            except ClientError as e:
                if e.response['Error']['Code'] == "404":
                    S3.create_bucket(Bucket=bucket)
                elif e.response['Error']['Code'] == "403":
                    raise Exception(("Incorrect AWS Access Key ID or Secret Access Key")) from e
                else:
                    raise
            except ConnectionError as e:
                raise Exception(("Cannot connect to Amazon S3")) from e
            if args.prep_memoize_dir[-1] == "/":
                align_memoize_dir = '%saln' % args.prep_memoize_dir
                align_stat_memoize_dir = '%sstats' % args.prep_memoize_dir
            else:
                align_memoize_dir = '%s/aln' % args.prep_memoize_dir
                align_stat_memoize_dir = '%s/stats' % args.prep_memoize_dir
        else:
            if not os.path.isdir(args.prep_memoize_dir):
                raise Exception(("Path '%s' does not exist") %
                    args.prep_memoize_dir)
            align_memoize_dir = os.path.join(args.prep_memoize_dir, 'aln')
            if not os.path.exists(align_memoize_dir):
                os.makedirs(align_memoize_dir)
            align_stat_memoize_dir = os.path.join(args.prep_memoize_dir, 'stats')
            if not os.path.exists(align_stat_memoize_dir):
                os.makedirs(align_stat_memoize_dir)

        am = align.AlignmentMemoizer(align_memoize_dir,
            aws_access_key_id=args.aws_access_key_id,
            aws_secret_access_key=args.aws_secret_access_key)
        asm = align.AlignmentStatMemoizer(align_stat_memoize_dir,
            aws_access_key_id=args.aws_access_key_id,
            aws_secret_access_key=args.aws_secret_access_key)
    else:
        am = None
        asm = None

    # Read list of taxonomies
    if args.input_type == 'auto-from-args':
        s = None if args.segment == 'None' else args.segment
        ref_accs = ncbi_neighbors.construct_references(args.tax_id) \
            if not args.ref_accs else args.ref_accs
        meta_filt = None
        meta_filt_against = None
        if args.metadata_filter:
            meta_filt = seq_io.read_metadata_filters(args.metadata_filter)
        if args.specific_against_metadata_filter:
            meta_filt_against = seq_io.read_metadata_filters(args.specific_against_metadata_filter)
        taxs = [(None, args.tax_id, s, ref_accs, meta_filt, meta_filt_against)]
    elif args.input_type == 'auto-from-file':
        taxs = seq_io.read_taxonomies(args.in_tsv)
    else:
        raise Exception(("Unknown input type '%s'") % args.input_type)

    # Read specified accessions, if provided
    if args.use_accessions:
        accessions_to_use = seq_io.read_accessions_for_taxonomies(
                args.use_accessions)
    else:
        accessions_to_use = None

    # Read specified sequences, if provided
    if args.use_fasta:
        sequences_to_use = seq_io.read_sequences_for_taxonomies(
                args.use_fasta)
    else:
        sequences_to_use = None

    # Only design for certain taxonomies, if provided
    if args.only_design_for:
        taxs_to_design_for = seq_io.read_taxonomies_to_design_for(
                args.only_design_for)
    else:
        taxs_to_design_for = None

    # Construct alignments for each taxonomy
    in_fasta = []
    taxid_for_fasta = []
    years_tsv_per_aln = []
    aln_tmp_dirs = []
    out_tsv = []
    design_for = []
    specific_against_metadata_accs = []
    for label, tax_id, segment, ref_accs, meta_filt, meta_filt_against in taxs:
        aln_file_dir = tempfile.TemporaryDirectory()
        if args.cover_by_year_decay:
            years_tsv_tmp = tempfile.NamedTemporaryFile()
            years_tsv_tmp_name = years_tsv_tmp.name
        else:
            years_tsv_tmp = None
            years_tsv_tmp_name = None

        if accessions_to_use is not None:
            if (tax_id, segment) in accessions_to_use:
                accessions_to_use_for_tax = accessions_to_use[(tax_id, segment)]
            else:
                accessions_to_use_for_tax = None
        else:
            accessions_to_use_for_tax = None

        if sequences_to_use is not None:
            if (tax_id, segment) in sequences_to_use:
                sequences_to_use_for_tax = sequences_to_use[(tax_id, segment)]
            else:
                sequences_to_use_for_tax = None
        else:
            sequences_to_use_for_tax = None

        if (accessions_to_use_for_tax is not None and
                sequences_to_use_for_tax is not None):
            raise Exception(("Cannot use both --use-accessions and "
                "--use-fasta for the same taxonomy"))

        nc, specific_against_metadata_acc = prepare_alignment.prepare_for(
            tax_id, segment, ref_accs,
            aln_file_dir.name, aln_memoizer=am, aln_stat_memoizer=asm,
            sample_seqs=args.sample_seqs,
            prep_influenza=args.prep_influenza,
            years_tsv=years_tsv_tmp_name,
            cluster_threshold=args.cluster_threshold,
            accessions_to_use=accessions_to_use_for_tax,
            sequences_to_use=sequences_to_use_for_tax,
            meta_filt=meta_filt,
            meta_filt_against=meta_filt_against)

        for i in range(nc):
            in_fasta += [os.path.join(aln_file_dir.name, str(i) + '.fasta')]
            taxid_for_fasta += [tax_id]
            specific_against_metadata_accs.append(specific_against_metadata_acc)
            if taxs_to_design_for is None:
                design_for += [True]
            else:
                design_for += [(tax_id, segment) in taxs_to_design_for]
        years_tsv_per_aln += [years_tsv_tmp]
        aln_tmp_dirs += [aln_file_dir]

        if label is None:
            out_tsv += [args.out_tsv + '.' + str(i) for i in range(nc)]
        else:
            for i in range(nc):
                out_name = label + '.' + str(i) + '.tsv'
                out_tsv += [os.path.join(args.out_tsv_dir, out_name)]

        if args.write_input_seqs:
            # Write the sequences that are in the alignment being used
            # as input
            all_seq_names = []
            for i in range(nc):
                fn = os.path.join(aln_file_dir.name, str(i) + '.fasta')
                seqs = seq_io.read_fasta(fn)
                all_seq_names += list(seqs.keys())
            all_seq_names = sorted(all_seq_names)
            if label is None:
                # args.write_input_seqs gives the path to where to write
                # the list
                out_file = args.write_input_seqs
            else:
                # Determine where to write the sequence names based on
                # the label and args.out_tsv_dir
                out_name = label + '.input-sequences.txt'
                out_file = os.path.join(args.out_tsv_dir, out_name)
            with open(out_file, 'w') as fw:
                for name in all_seq_names:
                    fw.write(name + '\n')
        if args.write_input_aln:
            # Write the alignments being used as input
            for i in range(nc):
                fn = os.path.join(aln_file_dir.name, str(i) + '.fasta')
                if label is None:
                    # args.write_input_aln gives the prefix of the path to
                    # which to write the alignment
                    copy_path = args.write_input_aln + '.' + str(i)
                else:
                    # Determine where to write the alignment based on the
                    # label and args.out_tsv_dir
                    out_name = label + '.' + str(i) + '.fasta'
                    copy_path = os.path.join(args.out_tsv_dir, out_name)
                shutil.copyfile(fn, copy_path)

    # Combine all years tsv (there is one per fasta file)
    if any(f is not None for f in years_tsv_per_aln):
        years_tsv = tempfile.NamedTemporaryFile()
        with open(years_tsv.name, 'w') as fw:
            for tf in years_tsv_per_aln:
                if tf is not None:
                    with open(tf.name) as fin:
                        for line in fin:
                            fw.write(line)
                    tf.close()
    else:
        years_tsv = None

    return in_fasta, taxid_for_fasta, years_tsv, aln_tmp_dirs, out_tsv, design_for, specific_against_metadata_accs


def design_for_id(args):
    """Design guides for differential identification across targets.

    Args:
        args: namespace of arguments provided to this executable
    """
    # Create an alignment object for each input
    # If obj is not 'minimize-guides', guide_cover_frac and seq_groups
    #  will be set to None
    alns = []
    seq_groups_per_input = []
    guide_cover_frac_per_input = []
    primer_cover_frac_per_input = []
    for in_fasta in args.in_fasta:
        seqs = seq_io.read_fasta(in_fasta)
        if args.cover_by_year_decay:
            aln, seq_groups, cover_frac = seqs_grouped_by_year(seqs, args)
            if args.search_cmd == 'complete-targets':
                guide_cover_frac, primer_cover_frac = cover_frac
            else:
                guide_cover_frac = cover_frac
                primer_cover_frac = None
        else:
            aln = alignment.Alignment.from_list_of_seqs(list(seqs.values()))
            seq_groups = None
            guide_cover_frac = args.guide_cover_frac
            if args.search_cmd == 'complete-targets':
                primer_cover_frac = args.primer_cover_frac
            else:
                primer_cover_frac = None
        alns += [aln]
        seq_groups_per_input += [seq_groups]
        guide_cover_frac_per_input += [guide_cover_frac]
        primer_cover_frac_per_input += [primer_cover_frac]

    # Keep track of how many items in alns to use for design (the first N of them)
    num_aln_for_design = len(args.in_fasta)

    # Add sequence lists to alns to be specific against
    # Note that although these are being place in alns, they may not
    # be actual alignments (they can be unaligned sequences)
    # If specific-against-metadata-filter is specified, add metadata filtered accessions to alns
    # and store what index it is located at in alns. Also, store start and end indices of
    # specific_against_metadata sequence lists
    specific_against_metadata_indices = {}
    specific_against_metadata_start = len(alns)
    for i in range(num_aln_for_design):
        if len(args.specific_against_metadata_accs[i]) > 0:
            logger.info(("Fetching %d sequences within taxon %d to be specific against"),
                len(args.specific_against_metadata_accs[i]), args.taxid_for_fasta[i])
            seqs = prepare_alignment.fetch_sequences_for_acc_list(list(
                args.specific_against_metadata_accs[i]))
            seqs_list = alignment.SequenceList(list(seqs.values()))
            alns += [seqs_list]
            specific_against_metadata_indices[i] = len(alns) - 1
    specific_against_metadata_end = len(alns)
    # Also add the specific_against_fastas sequences into alns
    for specific_against_fasta in args.specific_against_fastas:
        seqs = seq_io.read_fasta(specific_against_fasta)
        seqs_list = alignment.SequenceList(list(seqs.values()))
        alns += [seqs_list]
    # Finally, add specific_against_taxa sequences into alns after
    # downloading them
    if args.specific_against_taxa:
        taxs_to_be_specific_against = seq_io.read_taxonomies_to_design_for(
                args.specific_against_taxa)
        for taxid, segment in taxs_to_be_specific_against:
            logger.info(("Fetching sequences to be specific against tax "
                "%d (segment: %s)"), taxid, segment)
            seqs = prepare_alignment.fetch_sequences_for_taxonomy(
                    taxid, segment)
            seqs_list = alignment.SequenceList(list(seqs.values()))
            alns += [seqs_list]

    specific_against_exists = ((len(args.specific_against_fastas) > 0 or
            args.specific_against_taxa is not None) or
            (specific_against_metadata_end - specific_against_metadata_start) > 0)

    required_guides, blacklisted_ranges, blacklisted_kmers = \
        parse_required_guides_and_blacklist(args)
    required_flanking_seqs = (args.require_flanking5, args.require_flanking3)

    # Allow G-U base pairing, unless it is explicitly disallowed
    allow_gu_pairs = not args.do_not_allow_gu_pairing

    # Assign an id in [0, 1, 2, ...] for each taxid to design for
    # Find all alignments with each taxid
    aln_with_taxid = defaultdict(set)
    for i, taxid in enumerate(args.taxid_for_fasta):
        aln_with_taxid[taxid].add(i)
    num_taxa = len(aln_with_taxid)
    logger.info(("Designing for %d taxa"), num_taxa)

    # Read taxonomies to ignore for specificity, if specified
    tax_ignore = {}
    if (args.input_type in ['auto-from-file', 'auto-from-args'] and
            args.taxa_to_ignore_for_specificity):
        tax_ignore = seq_io.read_taxonomy_specificity_ignore(
                args.taxa_to_ignore_for_specificity)
        if args.specific_against_fastas or args.specific_against_taxa:
            logger.warning(("Taxa to ignore for specificity cannot from "
                "--specific-against-*"))

    # Construct the data structure for guide queries to perform
    # differential identification
    if num_taxa > 1 or specific_against_exists:
        logger.info(("Constructing data structure to permit guide queries for "
            "differential identification"))
        if args.diff_id_method == "lshnn":
            aq = alignment_query.AlignmentQuerierWithLSHNearNeighbor(alns,
                    args.guide_length, args.diff_id_mismatches, allow_gu_pairs)
        elif args.diff_id_method == "shard":
            aq = alignment_query.AlignmentQuerierWithKmerSharding(alns,
                    args.guide_length, args.diff_id_mismatches, allow_gu_pairs)
        else:
            raise Exception(("Unknown method for querying specificity: '%s'" %
                args.diff_id_method))
        aq.setup()
    else:
        logger.info(("Only one taxon was provided, so not constructing "
            "data structure to permit queries for differential "
            "identification"))
        aq = None

    for i in range(num_aln_for_design):
        taxid = args.taxid_for_fasta[i]
        logger.info(("Finding guides for alignment %d (of %d), which is in "
            "taxon %d"), i + 1, num_aln_for_design, taxid)

        if args.design_for is not None and args.design_for[i] is False:
            logger.info("Skipping design for this alignment")
            continue

        aln = alns[i]
        seq_groups = seq_groups_per_input[i]
        guide_cover_frac = guide_cover_frac_per_input[i]
        primer_cover_frac = primer_cover_frac_per_input[i]
        required_guides_for_aln = required_guides[i]
        blacklisted_ranges_for_aln = blacklisted_ranges[i]
        alns_in_same_taxon = aln_with_taxid[taxid]
        # For metadata filtering, we only want to be specific against the
        # accessions in alns[specific_against_metadata_index]
        specific_against_metadata_index = specific_against_metadata_indices[i] \
            if i in specific_against_metadata_indices else None

        if aq is not None:
            guide_is_specific = aq.guide_is_specific_to_alns_fn(
                    alns_in_same_taxon, args.diff_id_frac,
                    do_not_memoize=args.do_not_memoize_guide_computations)
        else:
            # No specificity to check
            guide_is_specific = lambda guide: True

        def guide_is_suitable(guide):
            # Return True iff the guide does not contain a blacklisted
            # k-mer and is specific to aln

            # Return False if the guide contains a blacklisted k-mer
            for kmer in blacklisted_kmers:
                if kmer in guide:
                    return False

            # Return True if guide does not hit too many sequences in
            # alignments other than aln
            return guide_is_specific(guide)

        # Mask alignments from this taxon from being reported in queries
        # because we will likely get many guide sequences that hit its
        # alignments, but we do not care about these for checking specificity
        if aq is not None:
            for j in range(num_aln_for_design):
                if j in alns_in_same_taxon:
                    aq.mask_aln(j)
                    logger.info(("Masking alignment %d as it is taxon %d"),
                        j + 1, taxid)
            # Also mask any specific against metadata sequence lists that do not
            # match the one for this alignment
            if specific_against_metadata_index is not None:
                logger.info(("Masking all sequence lists for specificity except the one "
                    "for alignment %d (taxon %d)"), i + 1, taxid)

            for j in range(specific_against_metadata_start, specific_against_metadata_end):
                if specific_against_metadata_index != j:
                    aq.mask_aln(specific_against_metadata_index)

            # Also mask taxonomies to ignore when determining specificity
            # of taxid
            if taxid in tax_ignore:
                for ignore_taxid in tax_ignore[taxid]:
                    for j in aln_with_taxid[ignore_taxid]:
                        logger.info(("Masking alignment %d (from taxon %d) "
                            "from specificity queries"), j + 1, ignore_taxid)
                        aq.mask_aln(j)

        # Construct activity predictor
        if args.use_simple_binary_activity_prediction:
            predictor = predict_activity.SimpleBinaryPredictor(
                    args.guide_mismatches,
                    allow_gu_pairs,
                    required_flanking_seqs=required_flanking_seqs)
        elif (args.predict_activity_model_path or
                args.predict_cas13a_activity_model is not None):
            if args.predict_activity_model_path:
                cla_path, reg_path = args.predict_activity_model_path
            else:
                dir_path = get_project_path()
                cla_path_all = os.path.join(dir_path, 'models', 'classify',
                                        'cas13a')
                reg_path_all = os.path.join(dir_path, 'models', 'regress',
                                        'cas13a')
                if len(args.predict_cas13a_activity_model) not in (0,2):
                    raise Exception(("If setting versions for "
                        "--predict-cas13a-activity-model, both a version for "
                        "the classifier and the regressor must be set."))
                if (len(args.predict_cas13a_activity_model) == 0 or
                        args.predict_cas13a_activity_model[0] == 'latest'):
                    cla_version = get_latest_model_version(cla_path_all)
                else:
                    cla_version = args.predict_cas13a_activity_model[0]
                if (len(args.predict_cas13a_activity_model) == 0 or
                        args.predict_cas13a_activity_model[1] == 'latest'):
                    reg_version = get_latest_model_version(reg_path_all)
                else:
                    reg_version = args.predict_cas13a_activity_model[1]
                cla_path = os.path.join(cla_path_all, cla_version)
                reg_path = os.path.join(reg_path_all, reg_version)
            if args.predict_activity_thres:
                # Use specified thresholds on classification and regression
                cla_thres, reg_thres = args.predict_activity_thres
            else:
                # Use default thresholds specified with the model
                cla_thres, reg_thres = None, None
            predictor = predict_activity.Predictor(cla_path, reg_path,
                    classification_threshold=cla_thres,
                    regression_threshold=reg_thres)
        else:
            if args.predict_activity_thres:
                raise Exception(("Cannot set --predict-activity-thres without "
                    "setting --predict-activity-model-path or "
                    "--predict-cas13a-activity-model"))
            if args.obj == 'maximize-activity':
                raise Exception(("Either --predict-activity-model-path or "
                    "--predict-cas13a-activity-model must be specified if "
                    "--obj is 'maximize-activity' (unless "
                    "--use-simple-binary-activity-prediction is set)"))
            if args.predict_activity_degradation:
                raise Exception(("--predict-activity-model-path must be "
                    "specified if --predict-activity-degradation is set "
                    "(unless --use-simple-binary-activity-prediction is set)"))
            # Do not predict activity
            predictor = None

        # Construct mutator, if necessary
        if args.predict_activity_degradation:
            mutator = mutate.GTRSubstitutionMutator(aln, *args.predict_activity_degradation,
                           args.predict_activity_degradation_mu,
                           args.predict_activity_degradation_t,
                           args.predict_activity_degradation_n)
        else:
            mutator = None

        # Find an optimal set of guides for each window in the genome,
        # and write them to a file; ensure that the selected guides are
        # specific to this alignment
        if args.obj == 'minimize-guides':
            gs = guide_search.GuideSearcherMinimizeGuides(
                    aln,
                    args.guide_length,
                    args.guide_mismatches,
                    guide_cover_frac,
                    args.missing_thres,
                    seq_groups=seq_groups,
                    guide_is_suitable_fn=guide_is_suitable,
                    required_guides=required_guides_for_aln,
                    blacklisted_ranges=blacklisted_ranges_for_aln,
                    allow_gu_pairs=allow_gu_pairs,
                    required_flanking_seqs=required_flanking_seqs,
                    predictor=predictor,
                    do_not_memoize_guides=args.do_not_memoize_guide_computations)
        elif args.obj == 'maximize-activity':
            gs = guide_search.GuideSearcherMaximizeActivity(
                    aln,
                    args.guide_length,
                    args.soft_guide_constraint,
                    args.hard_guide_constraint,
                    args.penalty_strength,
                    args.missing_thres,
                    algorithm=args.maximization_algorithm,
                    guide_is_suitable_fn=guide_is_suitable,
                    required_guides=required_guides_for_aln,
                    blacklisted_ranges=blacklisted_ranges_for_aln,
                    allow_gu_pairs=allow_gu_pairs,
                    required_flanking_seqs=required_flanking_seqs,
                    predictor=predictor,
                    do_not_memoize_guides=args.do_not_memoize_guide_computations)

        if args.search_cmd == 'sliding-window':
            # Find an optimal set of guides for each window in the genome,
            # and write them to a file
            gs.find_guides_with_sliding_window(args.window_size,
                args.out_tsv[i],
                window_step=args.window_step,
                sort=args.sort_out,
                print_analysis=(args.log_level<=logging.INFO))
        elif args.search_cmd == 'complete-targets':
            # Find optimal targets (primer and guide set combinations),
            # and write them to a file
            if args.primer_gc_content_bounds is None:
                primer_gc_content_bounds = None
            else:
                primer_gc_content_bounds = tuple(args.primer_gc_content_bounds)
            ps = primer_search.PrimerSearcher(aln, args.primer_length,
                                              args.primer_mismatches,
                                              primer_cover_frac,
                                              args.missing_thres,
                                              seq_groups=seq_groups,
                                              primer_gc_content_bounds=primer_gc_content_bounds)

            if args.obj == 'minimize-guides':
                obj_type = 'min'
            elif args.obj == 'maximize-activity':
                obj_type = 'max'
            ts = target_search.TargetSearcher(ps, gs,
                obj_type=obj_type,
                max_primers_at_site=args.max_primers_at_site,
                max_target_length=args.max_target_length,
                obj_weights=args.obj_fn_weights,
                only_account_for_amplified_seqs=args.only_account_for_amplified_seqs,
                halt_early=args.halt_search_early, mutator=mutator)
            ts.find_and_write_targets(args.out_tsv[i],
                best_n=args.best_n_targets, no_overlap=args.do_not_overlap)
        else:
            raise Exception("Unknown search subcommand '%s'" % args.search_cmd)

        # i should no longer be masked from queries
        if aq is not None:
            aq.unmask_all_aln()


def run(args):
    logger = logging.getLogger(__name__)

    # Set random seed for entire program
    random.seed(args.seed)
    np.random.seed(args.seed)

    check_obj_args(args)

    logger.info("Running design.py with arguments: %s", args)

    # Set NCBI API key
    if args.input_type in ['auto-from-file', 'auto-from-args']:
        if args.ncbi_api_key:
            ncbi_neighbors.ncbi_api_key = args.ncbi_api_key

    if args.input_type in ['auto-from-file', 'auto-from-args']:
        if args.input_type == 'auto-from-file':
            if not os.path.isdir(args.out_tsv_dir):
                raise Exception(("Output directory '%s' does not exist") %
                    args.out_tsv_dir)

        # Prepare input alignments, stored in temp fasta files
        in_fasta, taxid_for_fasta, years_tsv, aln_tmp_dirs, out_tsv, design_for, specific_against_metadata_accs = prepare_alignments(args)
        args.in_fasta = in_fasta
        args.taxid_for_fasta = taxid_for_fasta
        args.out_tsv = out_tsv
        args.design_for = design_for
        args.specific_against_metadata_accs = specific_against_metadata_accs

        if args.cover_by_year_decay:
            # args.cover_by_year_decay contains two parameters: the year
            # with the highest cover and the decay; add in (to the beginning)
            # the file listing the years
            year_highest_cover, year_cover_decay = args.cover_by_year_decay
            args.cover_by_year_decay = (years_tsv.name, year_highest_cover,
                    year_cover_decay)
    elif args.input_type == 'fasta':
        if len(args.in_fasta) != len(args.out_tsv):
            raise Exception(("Number output TSVs must match number of input "
                "FASTAs"))
        args.design_for = None
        args.taxid_for_fasta = list(range(len(args.in_fasta)))
        args.specific_against_metadata_accs = [[] for _ in range(len(args.in_fasta))]
    else:
        raise Exception("Unknown input type subcommand '%s'" % args.input_type)

    design_for_id(args)

    # Close temporary files storing alignments
    if args.input_type in ['auto-from-file', 'auto-from-args']:
        for td in aln_tmp_dirs:
            td.cleanup()
        if years_tsv is not None:
            years_tsv.close()


def argv_to_args(argv):
    parser = argparse.ArgumentParser()

    ###########################################################################
    # OPTIONS AVAILABLE ACROSS ALL SUBCOMMANDS
    ###########################################################################
    base_subparser = argparse.ArgumentParser(add_help=False)

    # Guide length
    base_subparser.add_argument('-gl', '--guide-length', type=int, default=28,
        help="Length of guide to construct")

    # Objective function
    base_subparser.add_argument('--obj',
        choices=['maximize-activity', 'minimize-guides'],
        default='minimize-guides',
        help=(("Objective function to solve. 'maximize-activity' maximizes "
               "the expected activity of the guide set of the target genomes "
               "subject to soft and hard constraints on the size of the guide "
               "set. 'minimize-guides' minimizes the number of guides in the "
               "guide set subject to coverage constraints across the target "
               "genomes.")))

    ##########
    # Parameters for minimization objective

    # Number of guide mismatches
    base_subparser.add_argument('-gm', '--guide-mismatches', type=int,
        help=("Allow for this number of mismatches when "
              "determining whether a guide covers a sequence"))
    # Desired coverage of target sequences
    def check_cover_frac(val):
        fval = float(val)
        if fval > 0 and fval <= 1:
            # a float in (0, 1]
            return fval
        else:
            raise argparse.ArgumentTypeError("%s is an invalid -p value" % val)
    base_subparser.add_argument('-gp', '--guide-cover-frac',
        type=check_cover_frac,
        help=("The fraction of all sequences that must be covered "
              "by the guides."))
    # Automatically setting desired coverage of target sequences based
    # on their year
    class ParseCoverDecayWithYearsFile(argparse.Action):
        # This is needed because --cover-by-year-decay has multiple args
        # of different types
        def __call__(self, parser, namespace, values, option_string=None):
            a, b, c = values
            # Check that b is a valid year
            year_pattern = re.compile('^(\d{4})$')
            if year_pattern.match(b):
                bi = int(b)
            else:
                raise argparse.ArgumentTypeError(("%s is an invalid 4-digit "
                    "year") % b)
            # Check that c is a valid decay
            cf = float(c)
            if cf <= 0 or cf >= 1:
                raise argparse.ArgumentTypeError(("%s is an invalid decay; it "
                    "must be a float in (0,1)" % c))
            setattr(namespace, self.dest, (a, bi, cf))
    class ParseCoverDecayByGeneratingYearsFile(argparse.Action):
        # This is needed because --cover-by-year-decay has multiple args
        # of different types
        def __call__(self, parser, namespace, values, option_string=None):
            a, b = values
            # Check that a is a valid year
            year_pattern = re.compile('^(\d{4})$')
            if year_pattern.match(a):
                ai = int(a)
            else:
                raise argparse.ArgumentTypeError(("%s is an invalid 4-digit "
                    "year") % a)
            # Check that b is a valid decay
            bf = float(b)
            if bf <= 0 or bf >= 1:
                raise argparse.ArgumentTypeError(("%s is an invalid decay; it "
                    "must be a float in (0,1)" % b))
            setattr(namespace, self.dest, (ai, bf))
    ##########

    ##########
    # Parameters for maximization objective

    # Soft guide constraint
    base_subparser.add_argument('-sgc', '--soft-guide-constraint', type=int,
        help=("Soft constraint on the number of guides. There is no "
              "penalty for a number of guides <= SOFT_GUIDE_CONSTRAINT, "
              "and having a number of guides beyond this is penalized. "
              "See --penalty-strength. This value must be <= "
              "HARD_GUIDE_CONSTRAINT."))
    # Hard guide constraint
    base_subparser.add_argument('-hgc', '--hard-guide-constraint', type=int,
        help=("Hard constraint on the number of guides. The number of "
              "guides designed for a target will be <= "
              "HARD_GUIDE_CONSTRAINT."))
    # Penalty strength
    base_subparser.add_argument('--penalty-strength', type=float,
        help=("Importance of the penalty when the number of guides "
              "exceeds the soft guide constraint. Namely, for a guide "
              "set G, if the penalty strength is L and the soft "
              "guide constraint is h, then the penalty in the objective "
              "function is L*max(0, |G|-h). Must be >= 0. The value "
              "depends on the output of activity model and reflects a "
              "tolerance for more guides; for the default activity model "
              "reasonable values are in the range [0.1, 0.5]."))
    # Algorithm for solving
    base_subparser.add_argument('--maximization-algorithm',
        choices=['greedy', 'random-greedy'],
        help=("Algorithm to use for solving submodular maximization "
              "problem. 'greedy' is the canonical deterministic greedy "
              "algorithm (Nemhauser 1978) for constrained monotone submodular "
              "maximization, which may perform well in practice but has "
              "poor theoretical guarantees here because the function is "
              "not monotone (unless --penalty-strength is 0). 'random-"
              "greedy' is the randomized greedy algorithm (Buchbinder "
              "2014) for constrained non-monotone submodular maximization "
              "that has good worst-case theoretical guarantees."))
    ##########

    # Handling missing data
    base_subparser.add_argument('--missing-thres', nargs=3,
        type=float, default=[0.5, 0.05, 1.5],
        help=("<A> <B> <C>; parameters governing the threshold on which sites "
              "to ignore due to too much missing data. The 3 values specify "
              "not to attempt to design guides overlapping sites where the "
              "fraction of sequences with missing data is > min(A, max(B, C*m)) "
              "where m is the median fraction of sequences with missing data "
              "over the alignment. Set a=1 and b=1 to not ignore sites due "
              "to missing data."))

    # Differential identification
    base_subparser.add_argument('--id-m', dest="diff_id_mismatches",
        type=int, default=4,
        help=("Allow for this number of mismatches when determining whether "
              "a guide 'hits' a sequence in a group/taxon other than the "
              "for which it is being designed; higher values correspond to more "
              "specificity."))
    base_subparser.add_argument('--id-frac', dest="diff_id_frac",
        type=float, default=0.01,
        help=("Decide that a guide 'hits' a group/taxon if it 'hits' a "
              "fraction of sequences in that group/taxon that exceeds this "
              "value; lower values correspond to more specificity."))
    base_subparser.add_argument('--id-method', dest="diff_id_method",
        choices=["lshnn", "shard"], default="shard",
        help=("Choice of method to query for specificity. 'lshnn' for "
              "LSH near-neighbor approach. 'shard' for approach that "
              "shards k-mers across small tries."))
    base_subparser.add_argument('--specific-against-fastas', nargs='+',
        default=[],
        help=("Path to one or more FASTA files giving sequences, such that "
              "guides are designed to be specific against (i.e., not hit) "
              "these sequences, according to --id-m and --id-frac. This "
              "is equivalent to specifying the FASTAs in the main input "
              "(as positional inputs), except that, when provided here, "
              "guides are not designed for them and they do not "
              "need to be aligned."))
    base_subparser.add_argument('--specific-against-taxa',
        help=("Path to TSV file giving giving taxonomies from which to "
              "download all genomes and ensure guides are specific against "
              "(i.e., not hit) these. The TSV file has 2 columns: (1) a "
              "taxonomic ID; (2) segment label, or 'None' if unsegmented"))

    # G-U pairing options
    base_subparser.add_argument('--do-not-allow-gu-pairing', action='store_true',
        help=("When determining whether a guide binds to a region of "
              "target sequence, do not count G-U (wobble) base pairs as "
              "matching. Default is to tolerate G-U pairing: namely, "
              "A in an output guide sequence matches G in the "
              "target and C in an output guide sequence matches T "
              "in the target (since the synthesized guide is the reverse "
              "complement of the output guide sequence)"))

    # Requiring guides in the cover, and blacklisting ranges and/or k-mers
    base_subparser.add_argument('--required-guides',
        help=("Path to a file that gives guide sequences that will be "
              "included in the guide cover and output for the windows "
              "in which they belong, e.g., if certain guide sequences are "
              "shown experimentally to perform well. The file must have "
              "3 columns: col 1 gives an identifier for the alignment "
              "that the guide covers, such that i represents the i'th "
              "FASTA given as input (0-based); col 2 gives a guide sequence; "
              "col 3 gives the start position of the guide (0-based) in "
              "the alignment"))
    base_subparser.add_argument('--blacklisted-ranges',
        help=("Path to a file that gives ranges in alignments from which "
              "guides will not be constructed. The file must have 3 columns: "
              "col 1 gives an identifier for the alignment that the range "
              "corresponds to, such that i represents the i'th FASTA "
              "given as input (0-based); col 2 gives the start position of "
              "the range (inclusive); col 3 gives the end position of the "
              "range (exclusive)"))
    base_subparser.add_argument('--blacklisted-kmers',
        help=("Path to a FASTA file that gives k-mers to blacklisted from "
              "guide sequences. No guide sequences will be constructed that "
              "contain these k-mers. The k-mers make up the sequences in "
              "the FASTA file; the sequence names are ignored. k-mers "
              "should be long enough so that not too many guide sequences "
              "are deemed to be unsuitable, and should be at most the "
              "length of the guide"))

    # Requiring flanking sequence (PFS)
    base_subparser.add_argument('--require-flanking5',
        help=("Require the given sequence on the 5' protospacer flanking "
              "site (PFS) of each designed guide; this tolerates ambiguity "
              "in the sequence (e.g., 'H' requires 'A', 'C', or 'T', or, "
              "equivalently, avoids guides flanked by 'G'). Note that "
              "this is the 5' end in the target sequence (not the spacer "
              "sequence)."))
    base_subparser.add_argument('--require-flanking3',
        help=("Require the given sequence on the 3' protospacer flanking "
              "site (PFS) of each designed guide; this tolerates ambiguity "
              "in the sequence (e.g., 'H' requires 'A', 'C', or 'T', or, "
              "equivalently, avoids guides flanked by 'G'). Note that "
              "this is the 3' end in the target sequence (not the spacer "
              "sequence)."))

    # Use a model to predict activity
    base_subparser.add_argument('--predict-activity-model-path',
        nargs=2,
        help=("Paths to directories containing serialized models in "
              "TensorFlow's SavedModel format for predicting guide-target "
              "activity. There are two arguments: (1) classification "
              "model to determine which guides are active; (2) regression "
              "model, which is used to determine which guides (among "
              "active ones) are highly active. The models/ directory "
              "contains example models. If neither this nor "
              "predict-cas13a-activity-model is set, ADAPT does not "
              "predict activities to use during design."))
    base_subparser.add_argument('--predict-cas13a-activity-model',
        nargs='*',
        help=("If set, use ADAPT's premade Cas13a model to "
              "predict guide-target activity. If neither this nor "
              "predict-activity-model-path is set, ADAPT does not predict "
              "activities to use during design. Optionally, two arguments can "
              "be included to indicate version number, in the format 'v1_0' or "
              "'latest'. Versions will default to latest."))
    base_subparser.add_argument('--predict-activity-thres',
        type=float,
        nargs=2,
        help=("Thresholds to use for decisions on output of predictive "
            "models. There are two arguments: (1) classification threshold "
            "for deciding which guide-target pairs are active (in [0,1], "
            "where higher values have higher precision but less recall); "
            "(2) regression threshold for deciding which guide-target pairs "
            "are highly active (>= 0, where higher values limit the number "
            "determined to be highly active). If not set but --predict-"
            "activity-model-path or --predict-cas13a-activity-model is set, "
            "then ADAPT uses default thresholds stored with the models."))
    base_subparser.add_argument('--use-simple-binary-activity-prediction',
        action='store_true',
        help=("If set, predict activity using a simple binary prediction "
              "between guide and target according to their distance, with "
              "the threshold determined based on --guide-mismatches. This "
              "is only applicable when OBJ is 'maxmimize-activity'. This "
              "does not use a serialized model for predicting activity, so "
              "neither --predict-activity-model-path nor "
              "--predict-cas13a-activity-model should be set when this "
              "is set."))
    # Predict activity degradation
    base_subparser.add_argument('--predict-activity-degradation',
        nargs=6, type=float,
        help=("If set, predict the degradation in activity over time due to "
              "substitution using the General Time-Reversible model. Six "
              "arguments should be set: the relative rates of substitutions "
              "per year between (1) A and C, (2) A and G, (3) A and T, "
              "(4) C and G, (5) C and T, & (6) G and T. Each of these should "
              "be between 0 and 1. Base pair frequencies will be calculated "
              "from input sequences. The 5th percentile of simulated activities "
              "will be reported."))
    base_subparser.add_argument('--predict-activity-degradation-t',
        type=int, default=5,
        help=("Amount of time to simulate substitutions over, in years "
              "(defaults to 5)"))
    base_subparser.add_argument('--predict-activity-degradation-mu',
        type=float, default=0.001,
        help=("Overall rate of substitutions per site per year (defaults to 0.001)"))
    base_subparser.add_argument('--predict-activity-degradation-n',
        type=int, default=500,
        help=("Number of sequences to simulate mutations over (defaults "
              "to 500). Higher values increase accuracy, but also increase"
              "runtime."))

    # Technical options
    base_subparser.add_argument('--do-not-memoize-guide-computations',
        action='store_true',
        help=("If set, do not memoize computations during the search, "
              "including of guides identified at each site and of "
              "specificity queries. This can be helpful for benchmarking "
              "the improvement of memoization, or if there is reason "
              "to believe memoization will slow the search (e.g., "
              "if possible amplicons rarely overlap). Note that activity "
              "predictions are still memoized."))
    base_subparser.add_argument('--seed', type=int,
        help=("SEED will set the random seed, guaranteeing the same output "
              "given the same inputs. If SEED is not set to the same value, "
              "output may vary across different runs."))

    # Log levels
    base_subparser.add_argument("--debug",
        dest="log_level",
        action="store_const",
        const=logging.DEBUG,
        default=logging.WARNING,
        help=("Debug output"))
    base_subparser.add_argument("--verbose",
        dest="log_level",
        action="store_const",
        const=logging.INFO,
        help=("Verbose output"))
    ###########################################################################

    ###########################################################################
    # SUBCOMMANDS FOR SEARCH TYPE
    ###########################################################################
    search_subparsers = parser.add_subparsers(dest='search_cmd')

    # Subcommand: sliding-window
    parser_sw = search_subparsers.add_parser('sliding-window',
        help=("Search for guides within a sliding window of a fixed size, "
              "and output the optimal guide set for each window"))
    parser_sw_args = argparse.ArgumentParser(add_help=False)
    parser_sw_args.add_argument('-w', '--window-size', type=int, default=200,
        help=("Ensure that selected guides are all a "
              "window of this size"))
    parser_sw_args.add_argument('--window-step', type=int, default=1,
        help=("Amount by which to increase the window start position for "
              "every iteration"))
    parser_sw_args.add_argument('--sort', dest='sort_out', action='store_true',
        help=("If set, sort output TSV by number of guides "
              "(ascending) then by score (descending); "
              "default is to sort by window position"))

    # Subcommand: complete-targets
    parser_ct = search_subparsers.add_parser('complete-targets',
        help=("Search for primer pairs and guides between them. This "
              "outputs the best BEST_N_TARGETS according to a cost "
              "function, where each target contains primers that bound "
              "an amplicon and a guide set within that amplicon."))
    parser_ct_args = argparse.ArgumentParser(add_help=False)
    parser_ct_args.add_argument('-pl', '--primer-length', type=int, default=30,
        help=("Length of primer in nt"))
    parser_ct_args.add_argument('-pp', '--primer-cover-frac',
        type=check_cover_frac, default=1.0,
        help=("Same as --cover-frac, except for the design of primers -- "
              "i.e., the fraction of sequences that must be covered "
              "by the primers, independently on each end"))
    parser_ct_args.add_argument('-pm', '--primer-mismatches',
        type=int, default=0,
        help=("Allow for this number of mismatches when determining "
              "whether a primer hybridizes to a sequence"))
    parser_ct_args.add_argument('--max-primers-at-site', type=int,
        help=("Only use primer sites that contain at most this number "
              "of primers; if not set, there is no limit"))
    parser_ct_args.add_argument('--primer-gc-content-bounds',
            nargs=2, type=float,
        help=("Only use primer sites where all primers are within the "
              "given GC content bounds. This consists of two values L and H, "
              "each fractions in [0,1], such that primer GC content must be "
              "in [L, H]. If not set, there are no bounds."))
    parser_ct_args.add_argument('--max-target-length', type=int,
        help=("Only allow amplicons (incl. primers) to be at most this "
              "number of nucleotides long; if not set, there is no limit"))
    parser_ct_args.add_argument('--obj-fn-weights', type=float, nargs=2,
        help=("Specify custom weights to use in the objective function "
              "for a target. These specify weights for penalties on primers "
              "and amplicons relative to the guide objective. There are "
              "2 weights (A B), where the target objective function "
              "is [(guide objective value) +/- (A*(total number of "
              "primers) + B*log2(amplicon length)]. It is + when "
              "--obj is minimize-guides and - when --obj is "
              "maximize-activity."))
    parser_ct_args.add_argument('--best-n-targets', type=int, default=10,
        help=("Only compute and output up to this number of targets. Note "
              "that runtime will generally be longer for higher values"))
    parser_ct_args.add_argument('--halt-search-early',
        action='store_true',
        help=('If set, stop the target search as soon as BEST_N_TARGETS '
              'have been identified. The targets will meet the given '
              'constraints but may not be optimal over the whole genome. '
              'They will likely be from the beginning of the genome.'))
    parser_ct_args.add_argument('--only-account-for-amplified-seqs',
        action='store_true',
        help=("If set, design guides to cover GUIDE_COVER_FRAC of just "
              "the sequences covered by the primers. This changes the "
              "behavior of -gp/--guide-cover-frac. This is only "
              "applicable when --obj is 'minimize-guides' as it is "
              "not implemented for 'maximize-activity'. In total, "
              ">= (GUIDE_COVER_FRAC * (2 * PRIMER_COVER_FRAC - 1)) "
              "sequences will be covered. Using this may worsen runtime "
              "because the sequences to consider for guide design will "
              "change more often across amplicons and therefore designs "
              "can be less easily memoized."))
    parser_ct_args.add_argument('--do-not-overlap',
        choices=['amplicon', 'primer', 'none'],
        default='amplicon',
        help=("What should be prevented from overlapping in the outputted targets. "
              "'amplicon' (default) prevents overlapping amplicons (target ranges) "
              "'primer' prevents both primers from overlapping (allowing 1 primer "
              "overlap). 'none' allows targets to overlap. When not 'none', if a "
              "target overlaps with a target already in the output, this *replaces* "
              "the overlapping one with the new one if the new one has a better "
              "objective value. When set to 'none', many targets in the "
              "BEST_N_TARGETS may be very similar"))
    ###########################################################################

    ###########################################################################
    # SUBCOMMANDS FOR INPUT TYPE
    ###########################################################################
    search_cmd_parsers = [(parser_sw, parser_sw_args),
                          (parser_ct, parser_ct_args)]

    # FASTA input
    input_fasta_subparser = argparse.ArgumentParser(add_help=False)
    input_fasta_subparser.add_argument('in_fasta', nargs='+',
        help=("Path to input FASTA. More than one can be "
              "given for differential identification"))
    input_fasta_subparser.add_argument('-o', '--out-tsv',
        nargs='+', required=True,
        help=("Path to output TSV. If more than one input FASTA is given, the "
              "same number of output TSVs must be given; each output TSV "
              "corresponds to an input FASTA."))
    input_fasta_subparser.add_argument('--cover-by-year-decay', nargs=3,
        action=ParseCoverDecayWithYearsFile,
        help=("<A> <B> <C>; if set, group input sequences by year and set a "
              "desired partial cover for each year (fraction of sequences that "
              "must be covered by guides) as follows: A is a tsv giving "
              "a year for each input sequence (col 1 is sequence name "
              "matching that in the input FASTA, col 2 is year). All years "
              ">= A receive a desired cover fraction of GUIDE_COVER_FRAC "
              "for guides (and PRIMER_COVER_FRAC for primers). Each preceding "
              "year receives a desired cover fraction that decays by B -- "
              "i.e., year n is given B*(desired cover fraction of year n+1)."))

    # Auto prepare, common arguments
    input_auto_common_subparser = argparse.ArgumentParser(add_help=False)
    input_auto_common_subparser.add_argument('--mafft-path',
        required=True,
        help=("Path to mafft executable, used for generating alignments"))
    input_auto_common_subparser.add_argument('--prep-memoize-dir',
        help=("Path to directory in which to memoize alignments and "
              "statistics on them. If set to \"s3://BUCKET/PATH\", it "
              "will save to the S3 bucket if boto3 and botocore are "
              "installed and access key information exists via "
              "AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY or via AWS CLI. "
              "If not set, this does not memoize this information."))
    input_auto_common_subparser.add_argument('--sample-seqs', type=int,
        help=("After fetching accessions, randomly select SAMPLE_SEQS of them "
              "with replacement from each taxonomy any move forward "
              "in the design with these. This is useful for testing and "
              "measuring output growth as input size grows, as well as "
              "assessing the dispersion in output associated with some "
              "input sample."))
    input_auto_common_subparser.add_argument('--prep-influenza',
        action='store_true',
        help=("If set, fetch sequences using the NCBI Influenza database; "
              "should only be used for Influenza A or B virus taxonomies"))
    input_auto_common_subparser.add_argument('--cover-by-year-decay', nargs=2,
        action=ParseCoverDecayByGeneratingYearsFile,
        help=("<A> <B>; if set, group input sequences by year and set a "
              "desired partial cover for each year (fraction of sequences that "
              "must be covered by guides) as follows: All years "
              ">= A receive a desired cover fraction of GUIDE_COVER_FRAC "
              "for guides (and PRIMER_COVER_FRAC for primers). Each preceding "
              "year receives a desired cover fraction that decays by B -- "
              "i.e., year n is given B*(desired cover fraction of year n+1)."))
    input_auto_common_subparser.add_argument('--cluster-threshold',
        type=float,
        default=0.2,
        help=(("Maximum inter-cluster distance to use when clustering "
               "input sequences prior to alignment. Expressed as average "
               "nucleotide dissimilarity (1-ANI, where ANI is average "
               "nucleotide identity); higher values result in fewer "
               "clusters")))
    input_auto_common_subparser.add_argument('--use-accessions',
        help=("If set, use specified accessions instead of fetching neighbors "
              "for the given taxonomic ID(s). This provides a path to a TSV "
              "file with 3 columns: (1) a taxonomic ID; (2) segment label, "
              "or 'None' if unsegmented; (3) accession. Each row specifies "
              "an accession to use in the input, and values for columns 1 "
              "and 2 can appear in multiple rows."))
    input_auto_common_subparser.add_argument('--use-fasta',
        help=("If set, use sequences in fasta instead of fetching neighbors "
              "for the given taxonomic ID(s). This provides a path to a TSV "
              "file with 3 columns: (1) a taxonomic ID; (2) segment label, "
              "or 'None' if unsegmented; (3) path to FASTA."))
    input_auto_common_subparser.add_argument('--only-design-for',
        help=("If set, only design for given taxonomies. This provides a "
              "path to a TSV file with 2 columns: (1) a taxonomic ID; (2) "
              "segment label, or 'None' if unsegmented"))
    input_auto_common_subparser.add_argument('--taxa-to-ignore-for-specificity',
        help=("If set, specify which taxa should be ignored when "
              "enforcing specificity while designing for other taxa. "
              "This provides a path to a TSV file with 2 columns: "
              "(1) a taxonomic ID A; (2) a taxonomic ID B such that "
              "B should be ignored when determining specificity for A. "
              "When designing for A, this masks taxonomy B from all "
              "specificity queries. This is useful, e.g., if B is a "
              "subset of A."))
    input_auto_common_subparser.add_argument('--ncbi-api-key',
        help=("API key to use for NCBI e-utils. Using this increases the "
              "limit on requests/second and may prevent an IP address "
              "from being blocked due to too many requests"))
    input_auto_common_subparser.add_argument('--aws-access-key-id',
        help=("User Account Access Key for AWS. This is only necessary "
            "if using S3 for memoization via PREP_MEMOIZE_DIR and AWS CLI "
            "is not installed and configured."))
    input_auto_common_subparser.add_argument('--aws-secret-access-key',
        help=("User Account Secret Access Key for AWS. This is only "
            "necessary if using S3 for memoization via PREP_MEMOIZE_DIR "
            "and AWS CLI is not installed and configured."))

    # Auto prepare from file
    input_autofile_subparser = argparse.ArgumentParser(add_help=False)
    input_autofile_subparser.add_argument('in_tsv',
        help=("Path to input TSV. Each row gives the following columns, "
              "in order: (1) label for the row (used for naming output "
              "files; must be unique); (2) taxonomic (e.g., species) ID from "
              "NCBI; (3) label of segment (e.g., 'S') if there is one, or "
              "'None' if unsegmented; (4) accessions of reference sequences to "
              "use for curation (comma-separated)"))
    input_autofile_subparser.add_argument('out_tsv_dir',
        help=("Path to directory in which to place output TSVs; each "
              "output TSV corresponds to a cluster for the taxon in a row "
              "in the input"))
    input_autofile_subparser.add_argument('--write-input-seqs',
        action='store_true',
        help=("If set, write the sequences (accession.version) being used as "
              "input for design to a file in OUT_TSV_DIR; the filename is "
              "determined based on the label for each taxonomy"))
    input_autofile_subparser.add_argument('--write-input-aln',
        action='store_true',
        help=("If set, write the alignments being used as "
              "input for design to a file in OUT_TSV_DIR; the filename is "
              "determined based on the label for each taxonomy (they are "
              "'[label].[cluster-number].fasta'"))

    # Auto prepare from arguments
    input_autoargs_subparser = argparse.ArgumentParser(add_help=False)
    input_autoargs_subparser.add_argument('tax_id', type=int,
        help=("Taxonomic (e.g., species) ID from NCBI"))
    input_autoargs_subparser.add_argument('segment',
        help=("Label of segment (e.g., 'S') if there is one, or 'None' if "
              "unsegmented"))
    input_autoargs_subparser.add_argument('out_tsv',
        help=("Path to output TSVs, with one per cluster; output TSVs are "
              "OUT_TSV.{cluster-number}"))
    input_autoargs_subparser.add_argument('--ref-accs', nargs='+',
        help=("Accession(s) of reference sequence(s) to use for curation (space-"
              "separated). If not set, ADAPT will automatically get accessions "
              "for reference sequences from NCBI based on the taxonomic ID"))
    input_autoargs_subparser.add_argument('--metadata-filter', nargs='+',
        help=("Only include accessions of specified taxonomic ID that match this metadata "
            "in the design. Metadata options are year, taxid, and country. Format as "
            "'metadata=value' or 'metadata!=value'. Separate multiple values with commas "
            "and different metadata filters with spaces (e.g. '--metadata-filter "
            "year!=2020,2019 taxid=11060')"))
    input_autoargs_subparser.add_argument('--specific-against-metadata-filter', nargs='+',
        help=("Only include accessions of the specified taxonomic ID that do not match this "
            "metadata in the design, and be specific against any accession that does match "
            "this metadata. Metadata options are year, taxid, and country. Format as "
            "'metadata=value' or 'metadata!=value'. Separate multiple values with commas "
            "and different metadata filters with spaces (e.g. "
            "'--specific-against-metadata-filter year!=2020,2019 taxid=11060')"))
    input_autoargs_subparser.add_argument('--write-input-seqs',
        help=("Path to a file to which to write the sequences "
              "(accession.version) being used as input for design"))
    input_autoargs_subparser.add_argument('--write-input-aln',
        help=("Prefix of path to files to which to write the alignments "
              "being used as input for design; filenames are "
              "'WRITE_INPUT_ALN.[cluster-number]'"))

    # Add parsers for subcommands
    for search_cmd_parser, search_cmd_parser_args in search_cmd_parsers:
        parents = [base_subparser, search_cmd_parser_args]

        search_cmd_subparser = search_cmd_parser.add_subparsers(
            dest='input_type')
        search_cmd_subparser.add_parser('fasta',
            parents=parents + [input_fasta_subparser],
            help=("Search from a given alignment input as a FASTA file"))
        search_cmd_subparser.add_parser('auto-from-file',
            parents=parents + [input_auto_common_subparser, input_autofile_subparser],
            help=("Automatically fetch sequences for one or more "
                  "taxonomies, then curate and align each; use these "
                  "alignments as input. The information is provided in "
                  "a TSV file. Differential identification is performed "
                  "across the taxonomies."))
        search_cmd_subparser.add_parser('auto-from-args',
            parents=parents + [input_auto_common_subparser, input_autoargs_subparser],
            help=("Automatically fetch sequences for one taxonomy, then curate "
                  "and align them; use this alignment as input. The "
                  "taxonomy is provided as command-line arguments."))
    ###########################################################################

    args = parser.parse_args(argv[1:])

    # Handle missing positional arguments by printing a help message
    if len(argv) == 1:
        # No arguments
        parser.print_help()
        sys.exit(1)
    if len(argv) == 2:
        # Only one position argument (missing input type)
        if args.search_cmd == 'sliding-window':
            parser_sw.print_help()
        if args.search_cmd == 'complete-targets':
            parser_ct.print_help()
        sys.exit(1)

    # Setup the logger
    log.configure_logging(args.log_level)

    return args


if __name__ == "__main__":
    run(argv_to_args(sys.argv))



"""Tests for align module.
"""

from collections import OrderedDict
import tempfile
from os import unlink
import unittest
try:
    import boto3
    import botocore
    S3 = boto3.client()
    resource = S3.list_buckets()
except:
    cloud = False
else:
    cloud = True
    import random
    import warnings

from adapt.prepare import align

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TestAlignStats(unittest.TestCase):
    """Tests functions manipulating and computing stats on alignments.
    """

    def test_aln_identity(self):
        a = 'ATCGATGGAT--G'
        b = 'ATAGTAGGATAAG'
        self.assertEqual(align._aln_identity(a, b), 8.0/13)

    def test_collapse_consecutive_gaps_simple(self):
        a = 'ATC----GA-TAT'
        b = 'ATCATCGGAAT--'

        expected_a = 'ATC-GA-TA'
        expected_b = 'ATCAGAAT-'

        self.assertEqual(align._collapse_consecutive_gaps(a, b),
            (expected_a, expected_b))

    def test_collapse_consecutive_gaps_simple(self):
        a = 'ATC----GA'
        b = 'ATCATCGGA'

        expected_a = 'ATC-GA'
        expected_b = 'ATCAGA'

        self.assertEqual(align._collapse_consecutive_gaps(a, b),
            (expected_a, expected_b))

    def test_collapse_consecutive_gaps_complex(self):
        a = 'ATC----GA-TAT'
        b = 'ATCATCGGAAT--'

        expected_a = 'ATC-GA-TA'
        expected_b = 'ATCAGAAT-'

        self.assertEqual(align._collapse_consecutive_gaps(a, b),
            (expected_a, expected_b))


class TestIO(unittest.TestCase):
    """Tests i/o functions.
    """

    def setUp(self):
        # Create a temporary file
        self.fasta = tempfile.NamedTemporaryFile(mode='w', delete=False)
        self.fasta.write(">AB123.2 Name of the sequence\n")
        self.fasta.write("ATCG\n")
        self.fasta.write("\n")
        self.fasta.write(">KY456 Another sequence\n")
        self.fasta.write("AATTAA\n")
        self.fasta.write("\n")
        # Closes the file so that it can be reopened on Windows
        self.fasta.close()


    def test_read_unaligned_seqs(self):
        # Check that it keys by accession.version (or just accession)
        seqs = align.read_unaligned_seqs(self.fasta)
        self.assertEqual(len(seqs), 2)
        self.assertIn('AB123.2', seqs)
        self.assertEqual(seqs['AB123.2'], "ATCG")
        self.assertIn('KY456', seqs)
        self.assertEqual(seqs['KY456'], "AATTAA") 

    def tearDown(self):
        # Delete temporary file
        unlink(self.fasta.name)


class TestMemoization(unittest.TestCase):
    """Tests memoizers.
    """

    def setUp(self):
        self.tempdir = tempfile.TemporaryDirectory()
        if cloud:
            cloudtempfold = random.getrandbits(32) 
            s3 = boto3.resource("s3")
            self.bucket = s3.Bucket("%08x" % cloudtempfold)
            self.bucket.create()
            self.cloudtempdir = "s3://%08x" % cloudtempfold

    def test_alignment_memoizer(self):
        def test_on_dir(directory):
            # Create some fake alignments
            seqs1 = OrderedDict([('AB123.1', 'ATCG--A'),
                                 ('AB456.1', 'TTTTCCA')])
            seqs2 = OrderedDict([('KY123.1', 'AAAACCA'),
                                 ('AB456.1', 'TTTTCCA')])
            seqs3 = OrderedDict([('AB123.1', 'ATCG--A'),
                                 ('KY123.1', 'AAAACCA'),
                                 ('AB456.1', 'TTTTCCA')])
            seqs4 = OrderedDict([('AB123.1', 'ATCG--A'),
                                 ('YZ123.1', 'GGGGCCA'),
                                 ('DE456.1', 'CCCCCCA')])
            am = align.AlignmentMemoizer(directory)

            # Progressively save these and check the results
            am.save(seqs1)
            am.save(seqs2)
            self.assertEqual(am.get(seqs1.keys()), seqs1)
            self.assertEqual(am.get(seqs2.keys()), seqs2)
            self.assertIsNone(am.get(seqs3.keys()))
            self.assertIsNone(am.get(seqs4.keys()))

            am.save(seqs3)
            self.assertEqual(am.get(seqs1.keys()), seqs1)
            self.assertEqual(am.get(seqs2.keys()), seqs2)
            self.assertEqual(am.get(seqs3.keys()), seqs3)
            self.assertIsNone(am.get(seqs4.keys()))

            am.save(seqs4)
            self.assertEqual(am.get(seqs1.keys()), seqs1)
            self.assertEqual(am.get(seqs2.keys()), seqs2)
            self.assertEqual(am.get(seqs3.keys()), seqs3)
            self.assertEqual(am.get(seqs4.keys()), seqs4)

            seqs3 = OrderedDict([('AB123.1', 'ATCG--A'),
                                 ('KY123.1', 'GGGGGGG'),
                                 ('AB456.1', 'TTTTCCA')])
            am.save(seqs3)
            self.assertEqual(am.get(seqs1.keys()), seqs1)
            self.assertEqual(am.get(seqs2.keys()), seqs2)
            self.assertEqual(am.get(seqs3.keys()), seqs3)
            self.assertEqual(am.get(seqs4.keys()), seqs4)
            del am

            # Try reading from a new AlignmentMemoizer at the same path
            am_new = align.AlignmentMemoizer(directory)
            self.assertEqual(am_new.get(seqs1.keys()), seqs1)
            self.assertEqual(am_new.get(seqs2.keys()), seqs2)
            self.assertEqual(am_new.get(seqs3.keys()), seqs3)
            self.assertEqual(am_new.get(seqs4.keys()), seqs4)

            # Reorder keys and check that output is the same
            seqs4_keys = list(seqs4.keys())
            self.assertEqual(am_new.get(seqs4_keys), seqs4)
            self.assertEqual(am_new.get(seqs4_keys[::-1]), seqs4)
        
        test_on_dir(self.tempdir.name)
        
        # If Boto3 is installed, test AWS cloud services
        if cloud:
            test_on_dir(self.cloudtempdir)

    def test_alignment_stat_memoizer(self):
        def test_on_dir(directory):
            # Create some fake stats
            asm = align.AlignmentStatMemoizer(directory)
            asm.save(['AB123.1', 'KY456.2'], (0.8, 0.9))
            asm.save(['AB123.1', 'KY789.2'], (0.85, 0.95))
            self.assertEqual(asm.get(['AB123.1', 'KY456.2']), (0.8, 0.9))
            self.assertEqual(asm.get(['AB123.1', 'KY789.2']), (0.85, 0.95))
            self.assertEqual(asm.get(['KY789.2', 'AB123.1']), (0.85, 0.95))
            self.assertIsNone(asm.get(['KY456.2', 'KY789.2']))
            del asm

            # Check that a new AlignmentStatMemoizer can read these
            asm_new = align.AlignmentStatMemoizer(directory)
            self.assertEqual(asm_new.get(['AB123.1', 'KY456.2']), (0.8, 0.9))
            self.assertEqual(asm_new.get(['AB123.1', 'KY789.2']), (0.85, 0.95))
            self.assertEqual(asm_new.get(['KY789.2', 'AB123.1']), (0.85, 0.95))
            self.assertIsNone(asm_new.get(['KY456.2', 'KY789.2']))

            # Add a new stat to the memoizer
            asm_new.save(['AB123.1', 'KY000.1'], (0.5, 0.6))
            del asm_new

            # Check that we can read all the stats
            asm_new2 = align.AlignmentStatMemoizer(directory)
            self.assertEqual(asm_new2.get(['AB123.1', 'KY456.2']), (0.8, 0.9))
            self.assertEqual(asm_new2.get(['AB123.1', 'KY789.2']), (0.85, 0.95))
            self.assertEqual(asm_new2.get(['KY789.2', 'AB123.1']), (0.85, 0.95))
            self.assertIsNone(asm_new2.get(['KY456.2', 'KY789.2']))
            self.assertEqual(asm_new2.get(['AB123.1', 'KY000.1']), (0.5, 0.6))

        test_on_dir(self.tempdir.name)

        # If Boto3 is installed, test AWS cloud services
        if cloud:
            test_on_dir(self.cloudtempdir)

    def tearDown(self):
        self.tempdir.cleanup()
        if cloud:
            self.bucket.objects.delete()
            self.bucket.delete()


class TestCurateAgainstRef(unittest.TestCase):
    """Tests the curate_against_ref() function.
    """

    def setUp(self):
        self.seqs = {'AB123.1': 'ATCGAAATTTA',
                     'AB456.1': 'AGCGAAGTTA',
                     'AB789.1': 'GGGGG',
                     'KY123.1': 'GAAAA',
                     'KY456.1': 'TTCGAAATTTA',
                     'KY789.1': 'GAAATTT',
                     'KZ123.1': 'GGGGGGGGGGG'}

        aln1  = {'AB123.1': 'ATCGAAATTTA',
                 'AB456.1': 'AGCGAAGTT-A'}
        aln2  = {'AB123.1': 'ATCGAAATTTA',
                 'AB789.1': '---GGGGG---'}
        aln3  = {'AB123.1': 'ATCGAAATTTA',
                 'KY123.1': '---GAAAA---'}
        aln4  = {'AB123.1': 'ATCGAAATTTA',
                 'KY456.1': 'TTCGAAATTTA'}
        aln5  = {'AB123.1': 'ATCGAAATTTA',
                 'KY789.1': '---GAAATTT-'}
        aln6  = {'AB123.1': 'ATCGAAATTTA',
                 'KZ123.1': 'GGGGGGGGGGG'}
        aln7  = {'KZ123.1': 'GGGGGGGGGGG',
                 'AB456.1': 'AGCGAAGTTA-'}
        aln8  = {'KZ123.1': 'GGGGGGGGGGG',
                 'AB789.1': '---GGGGG---'}
        aln9  = {'KZ123.1': 'GGGGGGGGGGG',
                  'KY123.1': 'GAAAA------'}
        aln10 = {'KZ123.1': 'GGGGGGGGGGG',
                 'KY456.1': 'TTCGAAATTTA'}
        aln11 = {'KZ123.1': 'GGGGGGGGGGG',
                 'KY789.1': '--GAAATTT--'}
        alns = [aln1, aln2, aln3, aln4, aln5, aln6,
                aln7, aln8, aln9, aln10, aln11]

        expected_curated_seq_accs1 = ['AB123.1', 'AB456.1', 'KY456.1',
                'KY789.1']
        self.expected_curated_seqs1 = {acc: self.seqs[acc]
            for acc in expected_curated_seq_accs1}

        expected_curated_seq_accs2 = ['AB123.1', 'AB456.1', 'KY456.1',
                'KY789.1', 'KZ123.1']
        self.expected_curated_seqs2 = {acc: self.seqs[acc]
            for acc in expected_curated_seq_accs2}
        
        # Override align.align() to provide expected alignments, but
        # keep the real function
        self.align_real = align.align
        def a(s):
            for aln in alns:
                if s.keys() == aln.keys():
                    return aln
        align.align = a

    def test_ref_acc_with_ver(self):
        ref_acc = 'AB123.1'
        self.assertEqual(
            align.curate_against_ref(self.seqs, [ref_acc],
                remove_ref_accs=[]),
            self.expected_curated_seqs1)

    def test_ref_acc_without_ver(self):
        ref_acc = 'AB123'
        self.assertEqual(
            align.curate_against_ref(self.seqs, [ref_acc],
                remove_ref_accs=[]),
            self.expected_curated_seqs1)

    def test_multiple_ref_accs(self):
        ref_accs = ['KZ123', 'AB123']
        self.assertEqual(
            align.curate_against_ref(self.seqs, ref_accs,
                remove_ref_accs=[]),
            self.expected_curated_seqs2)

    def tearDown(self):
        # Reset align.align()
        align.align = self.align_real



"""Tests for trie module.
"""

import unittest

from adapt.utils import trie

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class LeafInfo(trie.LeafInfo):
    """Mock class (wrapping a list) to use for leaf_info.
    """
    def __init__(self, l=[]):
        self.l = l
    def extend(self, other):
        self.l.extend(other.l)
    def remove(self, x):
        if x in self.l:
            self.l.remove(x)
    def is_empty(self):
        return len(self.l) == 0
    def copy(self):
        return LeafInfo(list(self.l))
    def __contains__(self, x):
        return x in self.l
    def __eq__(self, other):
        return frozenset(tuple(self.l)) == frozenset(tuple(other.l))
    def __repr__(self):
        return str(self.l)
    def __hash__(self):
        return hash(frozenset(tuple(self.l)))


def all_3mers():
    r = []
    for c1 in ['A', 'C', 'G', 'T']:
        for c2 in ['A', 'C', 'G', 'T']:
            for c3 in ['A', 'C', 'G', 'T']:
                r += [c1 + c2 + c3]
    return r


class TestNode(unittest.TestCase):
    """Tests Node objects used in Trie.
    """


    def test_insert(self):
        r = trie.Node(None)
        r.insert('AC', LeafInfo([1]))
        r.insert('AC', LeafInfo([2]))
        r.insert('AT', LeafInfo([3]))
        r.insert('GG', LeafInfo([4]))

        # Root should only point to A and G edges (0 and 2)
        self.assertIsNone(r.leaf_info)
        self.assertIsNone(r.children[1])
        self.assertIsNone(r.children[3])

        child_A = r.children[0]
        child_G = r.children[2]

        # A should only point to C and T edges (1 and 3); G should only
        # point to G edge (2)
        self.assertIsNone(child_A.leaf_info)
        self.assertIsNone(child_G.leaf_info)
        self.assertIsNone(child_A.children[0])
        self.assertIsNone(child_A.children[2])
        self.assertIsNone(child_G.children[0])
        self.assertIsNone(child_G.children[1])
        self.assertIsNone(child_G.children[3])

        # Now check the leaves
        child_AC = child_A.children[1]
        child_AT = child_A.children[3]
        child_GG = child_G.children[2]
        self.assertEqual(child_AC.children, [None, None, None, None])
        self.assertEqual(child_AT.children, [None, None, None, None])
        self.assertEqual(child_GG.children, [None, None, None, None])
        self.assertEqual(child_AC.leaf_info, LeafInfo([1,2]))
        self.assertEqual(child_AT.leaf_info, LeafInfo([3]))
        self.assertEqual(child_GG.leaf_info, LeafInfo([4]))

    def test_all_children(self):
        r = trie.Node(None)
        r.insert('AAA', LeafInfo([1]))
        r.insert('GGG', LeafInfo([2]))
        r.insert('ACC', LeafInfo([3]))
        r.insert('ACG', LeafInfo([4]))
        r.insert('GTG', LeafInfo([5]))
        r.insert('ACC', LeafInfo([6]))

        root_children = r.all_children()
        self.assertCountEqual([c.char for c in root_children], ['A', 'G'])

        child_A_children = r.children[0].all_children()
        self.assertCountEqual([c.char for c in child_A_children], ['A', 'C'])

        child_G_children = r.children[2].all_children()
        self.assertCountEqual([c.char for c in child_G_children], ['G', 'T'])

    def test_children_of_without_gu(self):
        r = trie.Node(None)
        r.insert('AAA', LeafInfo([1]))
        r.insert('GGG', LeafInfo([2]))
        r.insert('ACC', LeafInfo([3]))
        r.insert('ACG', LeafInfo([4]))
        r.insert('GTG', LeafInfo([5]))
        r.insert('ACC', LeafInfo([6]))
        r.insert('TCC', LeafInfo([7]))

        # Check for 'AA' and 'AC'
        children_A = r.children_of('A', gu_pairing=False)
        self.assertCountEqual([c.char for c in children_A], ['A'])
        children_AA = children_A[0].children_of('A', gu_pairing=False)
        self.assertCountEqual([c.char for c in children_AA], ['A'])
        children_AC = children_A[0].children_of('C', gu_pairing=False)
        self.assertCountEqual([c.char for c in children_AC], ['C'])

        # There is no 'AG'
        children_AG = children_A[0].children_of('G', gu_pairing=False)
        self.assertCountEqual(children_AG, [])

    def test_children_of_with_gu(self):
        r = trie.Node(None)
        r.insert('AAA', LeafInfo([1]))
        r.insert('GGG', LeafInfo([2]))
        r.insert('ACC', LeafInfo([3]))
        r.insert('ACG', LeafInfo([4]))
        r.insert('GTG', LeafInfo([5]))
        r.insert('ACC', LeafInfo([6]))
        r.insert('TCC', LeafInfo([7]))

        # 'A' should match 'A' and 'G' with G-U pairing
        children_A = r.children_of('A', gu_pairing=True)
        self.assertCountEqual([c.char for c in children_A], ['A', 'G'])

    def test_len(self):
        r = trie.Node(None)
        r.insert('AAA', LeafInfo([1]))
        r.insert('GGG', LeafInfo([2]))
        r.insert('ACC', LeafInfo([3]))
        r.insert('ACG', LeafInfo([4]))
        r.insert('GTG', LeafInfo([5]))
        r.insert('ACC', LeafInfo([6]))
        r.insert('TCC', LeafInfo([7]))
        r.insert('ATC', LeafInfo([8]))

        # 'ACC' only gets counted once
        self.assertEqual(len(r), 7)

    def test_query_no_mismatches_without_gu(self):
        r = trie.Node(None)
        r.insert('AAA', LeafInfo([1]))
        r.insert('GGG', LeafInfo([2]))
        r.insert('ACC', LeafInfo([3]))
        r.insert('ACG', LeafInfo([4]))
        r.insert('GTG', LeafInfo([5]))
        r.insert('ACC', LeafInfo([6]))
        r.insert('TCC', LeafInfo([7]))

        true = {'AAA': LeafInfo([1]),
                'GGG': LeafInfo([2]),
                'ACC': LeafInfo([3,6]),
                'ACG': LeafInfo([4]),
                'GTG': LeafInfo([5]),
                'TCC': LeafInfo([7])}
        for query in all_3mers():
            if query in true:
                expected = [true[query]]
            else:
                expected = []
            self.assertCountEqual(r.query(query, 0, 0, False), expected)

    def test_query_no_mismatches_without_gu_with_insert_replace(self):
        r = trie.Node(None)
        r.insert('AAA', LeafInfo([1]), replace=True)
        r.insert('GGG', LeafInfo([2]), replace=True)
        r.insert('ACC', LeafInfo([3]), replace=True)
        r.insert('ACG', LeafInfo([4]), replace=True)
        r.insert('GTG', LeafInfo([5]), replace=True)
        r.insert('ACC', LeafInfo([6]), replace=True)
        r.insert('TCC', LeafInfo([7]), replace=True)
        r.insert('GTG', LeafInfo([8]), replace=True)
        r.insert('ACG', LeafInfo([9]))

        true = {'AAA': LeafInfo([1]),
                'GGG': LeafInfo([2]),
                'ACC': LeafInfo([6]),
                'ACG': LeafInfo([4,9]),
                'GTG': LeafInfo([8]),
                'TCC': LeafInfo([7])}
        for query in all_3mers():
            if query in true:
                expected = [true[query]]
            else:
                expected = []
            self.assertCountEqual(r.query(query, 0, 0, False), expected)

    def test_query_no_mismatches_with_gu(self):
        r = trie.Node(None)
        r.insert('AAA', LeafInfo([1]))
        r.insert('GGG', LeafInfo([2]))
        r.insert('ACC', LeafInfo([3]))
        r.insert('ACG', LeafInfo([4]))
        r.insert('GTG', LeafInfo([5]))
        r.insert('ACC', LeafInfo([6]))
        r.insert('TCC', LeafInfo([7]))
        r.insert('ATC', LeafInfo([8]))

        # 'A' can match 'A' or 'G'
        self.assertCountEqual(r.query('AAA', 0, 0, True),
                [LeafInfo([1]), LeafInfo([2])])
        self.assertCountEqual(r.query('ATG', 0, 0, True),
                [LeafInfo([5])])

        # 'G' can only match 'G'
        self.assertCountEqual(r.query('GGG', 0, 0, True),
                [LeafInfo([2])])

        # 'C' can match 'T'
        self.assertCountEqual(r.query('CCC', 0, 0, True),
                [LeafInfo([7])])
        self.assertCountEqual(r.query('ACC', 0, 0, True),
                [LeafInfo([3,6]), LeafInfo([8])])

    def test_query_1_mismatch_without_gu(self):
        r = trie.Node(None)
        r.insert('AAAAAAA', LeafInfo([1]))
        r.insert('AAATAAA', LeafInfo([2]))
        r.insert('AAATTAA', LeafInfo([3]))
        r.insert('CCCCCCC', LeafInfo([4]))
        r.insert('CCCGCCC', LeafInfo([5]))
        r.insert('CCCGGCC', LeafInfo([6]))
        r.insert('GGGGGGG', LeafInfo([7]))
        r.insert('GGGTGGG', LeafInfo([8]))
        r.insert('GGGTTGG', LeafInfo([9]))
        r.insert('TTTTTTT', LeafInfo([10]))
        r.insert('TTTGTTT', LeafInfo([11]))
        r.insert('TTTGGTT', LeafInfo([12]))

        # Test above keys with mismatches
        expected = {'AAAAAAA': [LeafInfo([1]), LeafInfo([2])],
                    'AAAGAAA': [LeafInfo([1]), LeafInfo([2])],
                    'CAAAAAA': [LeafInfo([1])],
                    'ATATATA': [],
                    'CCCCCCC': [LeafInfo([4]), LeafInfo([5])],
                    'CCCACCC': [LeafInfo([4]), LeafInfo([5])],
                    'GCCCCCC': [LeafInfo([4])],
                    'GGGGGGG': [LeafInfo([7]), LeafInfo([8])],
                    'GGGAGGG': [LeafInfo([7]), LeafInfo([8])],
                    'AGGGGGG': [LeafInfo([7])],
                    'TTTTTTT': [LeafInfo([10]), LeafInfo([11])],
                    'TTTATTT': [LeafInfo([10]), LeafInfo([11])],
                    'ATTTTTT': [LeafInfo([10])]}
        for query, expected_value in expected.items():
            self.assertCountEqual(r.query(query, 0, 1, False), expected_value)

    def test_query_1_mismatch_with_gu(self):
        r = trie.Node(None)
        r.insert('AAAAAAA', LeafInfo([1]))
        r.insert('AAATAAA', LeafInfo([2]))
        r.insert('AAATTAA', LeafInfo([3]))
        r.insert('CCCCCCC', LeafInfo([4]))
        r.insert('CCCGCCC', LeafInfo([5]))
        r.insert('CCCGGCC', LeafInfo([6]))
        r.insert('GGGGGGG', LeafInfo([7]))
        r.insert('GGGTGGG', LeafInfo([8]))
        r.insert('GGGTTGG', LeafInfo([9]))
        r.insert('TTTTTTT', LeafInfo([10]))
        r.insert('TTTGTTT', LeafInfo([11]))
        r.insert('TTTGGTT', LeafInfo([12]))

        # Test above keys with mismatches and allowing G-U pairing
        expected = {'AAAAAAA': [LeafInfo([1]), LeafInfo([2]), LeafInfo([7]), LeafInfo([8])],
                    'AAAGAAA': [LeafInfo([1]), LeafInfo([2]), LeafInfo([7]), LeafInfo([8])],
                    'CAAAAAA': [LeafInfo([1]), LeafInfo([7])],
                    'ATATATA': [],
                    'CCCCCCC': [LeafInfo([4]), LeafInfo([5]), LeafInfo([10]), LeafInfo([11])],
                    'CCCACCC': [LeafInfo([4]), LeafInfo([5]), LeafInfo([6]), LeafInfo([10]), LeafInfo([11]), LeafInfo([12])],
                    'GCCCCCC': [LeafInfo([4]), LeafInfo([10])],
                    'GGGGGGG': [LeafInfo([7]), LeafInfo([8])],
                    'GGGAGGG': [LeafInfo([7]), LeafInfo([8])],
                    'AGGGGGG': [LeafInfo([7]), LeafInfo([8])],
                    'TTTTTTT': [LeafInfo([10]), LeafInfo([11])],
                    'TTTATTT': [LeafInfo([10]), LeafInfo([11]), LeafInfo([12])],
                    'ATTTTTT': [LeafInfo([10])]}
        for query, expected_value in expected.items():
            self.assertCountEqual(r.query(query, 0, 1, True), expected_value)

    def test_remove_and_reinsert(self):
        r = trie.Node(None)
        r.insert('AAA', LeafInfo([1]))
        r.insert('GGG', LeafInfo([2]))
        r.insert('ACC', LeafInfo([3]))
        r.insert('ACG', LeafInfo([4]))
        r.insert('GTG', LeafInfo([5]))
        r.insert('ACC', LeafInfo([6]))
        r.insert('TCC', LeafInfo([7]))

        # Check if the contents are as expected
        true = {'AAA': LeafInfo([1]),
                'GGG': LeafInfo([2]),
                'ACC': LeafInfo([3,6]),
                'ACG': LeafInfo([4]),
                'GTG': LeafInfo([5]),
                'TCC': LeafInfo([7])}
        for query in all_3mers():
            if query in true:
                expected = [true[query]]
            else:
                expected = []
            self.assertCountEqual(r.query(query, 0, 0, False), expected)

        # Remove a few strings
        self.assertTrue(r.remove('GGG'))
        self.assertTrue(r.remove('ACC'))
        self.assertTrue(r.remove('GTG'))
        self.assertFalse(r.remove('GTG'))
        self.assertFalse(r.remove('CCC'))

        # Check if the contents are as expected
        true_with_removals = {'AAA': LeafInfo([1]),
                              'ACG': LeafInfo([4]),
                              'TCC': LeafInfo([7])}
        for query in all_3mers():
            if query in true_with_removals:
                expected = [true_with_removals[query]]
            else:
                expected = []
            self.assertCountEqual(r.query(query, 0, 0, False), expected)

        # Re-insert the removed strings
        r.insert('GGG', LeafInfo([2]))
        r.insert('ACC', LeafInfo([3]))
        r.insert('GTG', LeafInfo([5]))
        r.insert('ACC', LeafInfo([6]))

        # Check if the contents are as expected
        for query in all_3mers():
            if query in true:
                expected = [true[query]]
            else:
                expected = []
            self.assertCountEqual(r.query(query, 0, 0, False), expected)

    def test_traverse_and_find(self):
        r = trie.Node(None)
        r.insert('AAA', LeafInfo([1]))
        r.insert('GGG', LeafInfo([2]))
        r.insert('ACC', LeafInfo([3]))
        r.insert('ACG', LeafInfo([4]))
        r.insert('GTG', LeafInfo([5]))
        r.insert('ACC', LeafInfo([6]))
        r.insert('TCC', LeafInfo([7]))
        r.insert('GAA', LeafInfo([2]))
        r.insert('GCC', LeafInfo([5]))
        r.insert('GCC', LeafInfo([7]))

        self.assertCountEqual(r.traverse_and_find(1),
            [('AAA', LeafInfo([1]))])
        self.assertCountEqual(r.traverse_and_find(2),
            [('GGG', LeafInfo([2])), ('GAA', LeafInfo([2]))])
        self.assertCountEqual(r.traverse_and_find(3),
            [('ACC', LeafInfo([3,6]))])
        self.assertCountEqual(r.traverse_and_find(4),
            [('ACG', LeafInfo([4]))])
        self.assertCountEqual(r.traverse_and_find(5),
            [('GTG', LeafInfo([5])), ('GCC', LeafInfo([5,7]))])
        self.assertCountEqual(r.traverse_and_find(6),
            [('ACC', LeafInfo([3,6]))])
        self.assertCountEqual(r.traverse_and_find(7),
            [('TCC', LeafInfo([7])), ('GCC', LeafInfo([5,7]))])

    def test_traverse_and_remove(self):
        r = trie.Node(None)
        r.insert('AAA', LeafInfo([1]))
        r.insert('GGG', LeafInfo([2]))
        r.insert('ACC', LeafInfo([3]))
        r.insert('ACG', LeafInfo([4]))
        r.insert('GTG', LeafInfo([5]))
        r.insert('ACC', LeafInfo([6]))
        r.insert('TCC', LeafInfo([7]))
        r.insert('GAA', LeafInfo([2]))
        r.insert('GCC', LeafInfo([5]))
        r.insert('GCC', LeafInfo([7]))

        true = {'AAA': LeafInfo([1]),
                'GGG': LeafInfo([2]),
                'ACC': LeafInfo([3,6]),
                'ACG': LeafInfo([4]),
                'GTG': LeafInfo([5]),
                'TCC': LeafInfo([7]),
                'GAA': LeafInfo([2]),
                'GCC': LeafInfo([5,7])}
        def check_contents():
            # Check if the contents are as expected
            for query in all_3mers():
                if query in true:
                    expected = [true[query]]
                else:
                    expected = []
                self.assertCountEqual(r.query(query, 0, 0, False), expected)

        # Start by checking contents
        check_contents()

        # Repeatedly remove and check contents
        r.traverse_and_remove(3)
        true['ACC'] = LeafInfo([6])
        check_contents()

        r.traverse_and_remove(100)
        check_contents()

        r.traverse_and_remove(3)
        check_contents()

        r.traverse_and_remove(1)
        del true['AAA']
        check_contents()

        r.traverse_and_remove(2)
        del true['GGG']
        del true['GAA']
        check_contents()

        r.traverse_and_remove(5)
        true['GCC'] = LeafInfo([7])
        del true['GTG']
        check_contents()

        r.traverse_and_remove(7)
        del true['TCC']
        del true['GCC']
        check_contents()

        r.traverse_and_remove(4)
        del true['ACG']
        check_contents()

        r.traverse_and_remove(3)
        check_contents()

        r.traverse_and_remove(6)
        del true['ACC']
        check_contents()

        r.traverse_and_remove(1)
        check_contents()


class TestTrie(unittest.TestCase):
    """Tests methods in Trie class.

    Many of these operations are wrappers around methods in Node,
    so these unit tests are less extensive than ones in TestNode.
    """

    def test_insert_and_query(self):
        t = trie.Trie()
        t.insert([('AAA', LeafInfo([1])),
                  ('AAA', LeafInfo([1])),
                  ('GGG', LeafInfo([2])),
                  ('ACC', LeafInfo([3])),
                  ('ACG', LeafInfo([4])),
                  ('GTG', LeafInfo([5])),
                  ('ACC', LeafInfo([6])),
                  ('TCC', LeafInfo([7]))])

        true = {'AAA': LeafInfo([1]),
                'GGG': LeafInfo([2]),
                'ACC': LeafInfo([3,6]),
                'ACG': LeafInfo([4]),
                'GTG': LeafInfo([5]),
                'TCC': LeafInfo([7])}
        for query in all_3mers():
            if query in true:
                expected = [true[query]]
            else:
                expected = []
            self.assertCountEqual(t.query(query, 0, False), expected)

    def test_mask_and_unmask(self):
        t = trie.Trie()

        t.insert([('AAA', LeafInfo([1])),
                  ('GGG', LeafInfo([2])),
                  ('ACC', LeafInfo([3])),
                  ('ACG', LeafInfo([4])),
                  ('GTG', LeafInfo([5])),
                  ('ACC', LeafInfo([6])),
                  ('TCC', LeafInfo([7])),
                  ('GAA', LeafInfo([2])),
                  ('GCC', LeafInfo([5])),
                  ('GCC', LeafInfo([7]))])

        true = {'AAA': LeafInfo([1]),
                'GGG': LeafInfo([2]),
                'ACC': LeafInfo([3,6]),
                'ACG': LeafInfo([4]),
                'GTG': LeafInfo([5]),
                'TCC': LeafInfo([7]),
                'GAA': LeafInfo([2]),
                'GCC': LeafInfo([5,7])}
        true_orig = dict(true)
        def check_contents():
            # Check if the contents are as expected
            for query in all_3mers():
                if query in true:
                    expected = [true[query]]
                else:
                    expected = []
                self.assertCountEqual(t.query(query, 0, False), expected)

        # Start by checking contents
        check_contents()

        # Repeatedly mask and check contents
        t.mask(3)
        true['ACC'] = LeafInfo([6])
        check_contents()

        t.mask(100)
        check_contents()

        t.mask(3)
        check_contents()

        t.mask(1)
        del true['AAA']
        check_contents()

        t.mask(2)
        del true['GGG']
        del true['GAA']
        check_contents()

        t.mask(5)
        true['GCC'] = LeafInfo([7])
        del true['GTG']
        check_contents()

        t.mask(7)
        del true['TCC']
        del true['GCC']
        check_contents()

        t.mask(4)
        del true['ACG']
        check_contents()

        t.mask(3)
        check_contents()

        t.mask(6)
        del true['ACC']
        check_contents()

        t.mask(1)
        check_contents()

        # Unmask everything
        t.unmask_all()

        # Check contents
        true = true_orig
        check_contents()



"""Methods for searching for optimal primers through a genome.

This makes heavy use of the guide_search module.
"""

import logging

from adapt import guide_search
from adapt.utils import guide

__author__ = 'Hayden Metsky <hayden@mit.edu>'

logger = logging.getLogger(__name__)


class PrimerResult:
    """Store results of a primer cover at a site."""

    def __init__(self, start, num_primers, primer_length,
                 frac_bound, primers_in_cover):
        """
        Args:
            start: start position of the primer
            num_primers: number of primers designed at the site
            primer_length: length of each primer
            frac_bound: total fraction of all sequences bound by the primers
            primers_in_cover: set of primers that achieve the desired coverage
                 and is minimal at the site
        """
        self.start = start
        self.num_primers = num_primers
        self.primer_length = primer_length
        self.frac_bound = frac_bound
        self.primers_in_cover = primers_in_cover

    def overlaps(self, other, expand=0):
        """Determine if self overlaps other.

        Args:
            other: PrimerResult object
            expand: tests overlap within +/- EXPAND nt of other

        Returns:
            True iff self overlaps other
        """
        return ((self.start - expand <= other.start < (self.start +
                    self.primer_length + expand)) or
                (other.start - expand <= self.start < (other.start +
                    other.primer_length + expand)))

    def overlaps_range(self, start, end):
        """Determine if self overlaps a range.

        Args:
            start: start of range (inclusive)
            end: end of range (exclusive)

        Returns:
            True iff self overlaps (start, end)
        """
        return ((self.start < end) and (start < self.start + self.primer_length))

    def __str__(self):
        return str((self.start, self.num_primers, self.primer_length,
            self.frac_bound, self.primers_in_cover))

    def __repr__(self):
        return str((self.start, self.num_primers, self.primer_length,
            self.frac_bound, self.primers_in_cover))

    def __eq__(self, other):
        """Determine equality of self and other.

        Args:
            other: an object of PrimerResult

        Returns:
            True iff self is identical to other
        """
        return (self.start == other.start and
                self.num_primers == other.num_primers and
                self.primer_length == other.primer_length and
                self.frac_bound == other.frac_bound and
                self.primers_in_cover == other.primers_in_cover)


class PrimerSearcher(guide_search.GuideSearcherMinimizeGuides):
    """Methods to search for primers over a genome.

    This is a special case of guide_search.GuideSearcherMinimizeGuides; thus, it
    is a subclass of guide_search.GuideSearcherMinimizeGuides. This effectively
    looks for guides (here, primers) within each window of size w
    where w is the length of a primer.

    The input is an alignment of sequences over which to search, as well as
    parameters defining the space of acceptable primers.
    """

    def __init__(self, aln, primer_length, mismatches, cover_frac,
                 missing_data_params, seq_groups=None,
                 primer_gc_content_bounds=None):
        """
        Args:
            aln: alignment.Alignment representing an alignment of sequences
            primer_length: length of the primer to construct
            mismatches: threshold on number of mismatches for determining whether
                a primer would hybridize to a target sequence
            cover_frac: fraction in (0, 1] of sequences that must be 'captured' by
                 a primer; see seq_groups
            missing_data_params: tuple (a, b, c) specifying to not attempt to
                design guides overlapping sites where the fraction of
                sequences with missing data is > min(a, max(b, c*m)), where m is
                the median fraction of sequences with missing data over the
                alignment
            seq_groups: dict that maps group ID to collection of sequences in
                that group. If set, cover_frac must also be a dict that maps
                group ID to the fraction of sequences in that group that
                must be 'captured' by a primer. If None, then do not divide
                the sequences into groups.
            primer_gc_content_bounds: a tuple (lo, hi) such that this only
                yields sites where all primers have a GC content fraction in
                [lo, hi]; or None for no bounds
        """
        super().__init__(aln, primer_length, mismatches,
                         cover_frac, missing_data_params,
                         seq_groups=seq_groups,
                         allow_gu_pairs=False,
                         do_not_memoize_guides=True)
        self.primer_gc_content_bounds = primer_gc_content_bounds

    def seqs_bound_by_primers(self, primers):
        """Determine the sequences in the alignment bound by the primers.

        Args:
            primers: collection of str representing primer sequences

        Returns:
            set of sequence identifiers (index in alignment) bound by
            a primer
        """
        return super()._seqs_bound_by_guides(primers)

    def check_gc_content(self, primers):
        """Determine whether all primers in primer set meet bounds on GC content.

        Args:
            primers: collection of str representing primer sequences

        Returns:
            True/False indicating whether all primers meet the bounds on
            GC content
        """
        lo, hi = self.primer_gc_content_bounds
        assert lo <= hi
        for primer_seq in primers:
            gc_frac = guide.gc_frac(primer_seq)
            if gc_frac < lo or gc_frac > hi:
                return False
        return True

    def find_primers(self, max_at_site=None):
        """Find primers across the alignment.

        Args:
            max_at_site: only yield sites that have <= MAX_AT_SITE
                primers at a site; or None for no limit

        Yields:
            tuple at each site in the alignment, consisting of the
            following values, in order:
              1) start position of the primer
              2) number of primers designed at the site
              3) total fraction of all sequences bound by the primers
              4) set of primers that achieve the desired coverage
                 and is minimal at the site
            They are given in sorted order, by position in the
            alignment
        """
        window_size = self.guide_length # primer length
        for cover in self._find_guides_for_each_window(
                window_size, hide_warnings=True):
            start, end, primers_in_cover = cover
            num_primers = len(primers_in_cover)
            frac_bound = self.total_frac_bound_by_guides(primers_in_cover)

            # Check constraints
            if max_at_site is not None and num_primers > max_at_site:
                continue
            if (self.primer_gc_content_bounds is not None and
                    self.check_gc_content(primers_in_cover) is False):
                continue

            yield PrimerResult(
                start, num_primers, window_size,
                frac_bound, primers_in_cover)



"""Tests for alignment_query module.
"""

import random
import unittest

from adapt import alignment
from adapt.specificity import alignment_query

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class BaseAlignmentQuerierTests:
    """Tests the AlignmentQuerier classes.

    unittest won't run these tests directly because this does not inherit from
    unittest.TestCase.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        # Set a random seed so hash functions are always the same
        random.seed(0)

        aln_a_seqs = ['ATCGAAATAAACC',
                      'ATCGATATAATGG',
                      'ATGGTTATAATGG']
        aln_a = alignment.Alignment.from_list_of_seqs(aln_a_seqs)

        aln_b_seqs = ['GTTCCTTTACAGT',
                      'GCTCCTTCACAGT',
                      'GCTCCTTTCCCGT']
        aln_b = alignment.Alignment.from_list_of_seqs(aln_b_seqs)

        aln_c_seqs = ['ATCGACATAGTCG',
                      'ATCGACATAATGG',
                      'CTGGTCATAACCC']
        aln_c = alignment.Alignment.from_list_of_seqs(aln_c_seqs)

        alns = [aln_a, aln_b, aln_c]
        if self.alignment_querier_subclass == alignment_query.AlignmentQuerierWithLSHNearNeighbor:
            self.aq = self.alignment_querier_subclass(
                    alns, 5, 1, False, k=3, reporting_prob=0.95)
            self.aq_with_gu = self.alignment_querier_subclass(
                    alns, 5, 1, True, k=3, reporting_prob=0.95)
        else:
            self.aq = self.alignment_querier_subclass(
                    alns, 5, 1, False)
            self.aq_with_gu = self.alignment_querier_subclass(
                    alns, 5, 1, True)
        self.aq.setup()
        self.aq_with_gu.setup()

    def test_frac_of_aln_hit_by_guide(self):
        self.assertEqual(self.aq.frac_of_aln_hit_by_guide('ATCGA'),
                         [2.0/3, 0.0, 2.0/3])

        self.assertEqual(self.aq.frac_of_aln_hit_by_guide('GGGGG'),
                         [0.0, 0.0, 0.0])

        self.assertEqual(self.aq.frac_of_aln_hit_by_guide('CCTTC'),
                         [0.0, 1.0, 0.0])

        self.assertEqual(self.aq.frac_of_aln_hit_by_guide('TTACA'),
                         [1.0/3, 2.0/3, 0.0])

    def test_frac_of_aln_hit_by_guide_with_gu_pairs(self):
        # In aln_a, this also hits the 3rd sequence at 'AATGG'
        self.assertEqual(self.aq_with_gu.frac_of_aln_hit_by_guide('ATCGA'),
                         [1.0, 0.0, 2.0/3])

        self.assertEqual(self.aq_with_gu.frac_of_aln_hit_by_guide('GGGGG'),
                         [0.0, 0.0, 0.0])

        self.assertEqual(self.aq_with_gu.frac_of_aln_hit_by_guide('CCTTC'),
                         [0.0, 1.0, 0.0])

        self.assertEqual(self.aq_with_gu.frac_of_aln_hit_by_guide('TTACA'),
                         [2.0/3, 2.0/3, 2.0/3])

    def test_guide_is_specific_to_aln(self):
        def assert_is_specific(guide, thres, specific_to):
            not_specific_to = [i for i in [0, 1, 2] if i not in specific_to]
            for i in specific_to:
                self.assertTrue(self.aq.guide_is_specific_to_alns(guide, [i], thres))
            for i in not_specific_to:
                self.assertFalse(self.aq.guide_is_specific_to_alns(guide, [i], thres))

        assert_is_specific('ATCGA', 0, [])
        assert_is_specific('GGGGG', 0, [0, 1, 2])
        assert_is_specific('CCTTC', 0, [1])
        assert_is_specific('TTACA', 0, [])

        assert_is_specific('ATCGA', 0.5, [])
        assert_is_specific('GGGGG', 0.5, [0, 1, 2])
        assert_is_specific('CCTTC', 0.5, [1])
        assert_is_specific('TTACA', 0.5, [1])

    def test_hit_with_gaps(self):
        gappy_aln_seqs = ['ATCGACGTAAACC',
                          'ATCGA--TAATGG',
                          'ATGGA--TAATGG']
        gappy_aln = alignment.Alignment.from_list_of_seqs(gappy_aln_seqs)
        if self.alignment_querier_subclass == alignment_query.AlignmentQuerierWithLSHNearNeighbor:
            gappy_aq = self.alignment_querier_subclass(
                    [gappy_aln], 5, 1, False, k=3, reporting_prob=0.95)
        else:
            gappy_aq = self.alignment_querier_subclass(
                    [gappy_aln], 5, 1, False)
        gappy_aq.setup()

        # GGGGG should hit no sequences
        self.assertEqual(gappy_aq.frac_of_aln_hit_by_guide('GGGGG'),
                         [0.0])

        # GACGT should only hit the first sequence
        self.assertEqual(gappy_aq.frac_of_aln_hit_by_guide('GACGT'),
                         [1.0/3.0])

        # GATAA should hit the second and third sequence (since their
        # gaps are removed)
        self.assertEqual(gappy_aq.frac_of_aln_hit_by_guide('GATAA'),
                         [2.0/3.0])

        # CATAA should hit all 3 alignments (tolerating a mismatch
        # to CGTAA in the first, and with perfect match in the second
        # and third after their gaps are removed
        self.assertEqual(gappy_aq.frac_of_aln_hit_by_guide('CATAA'),
                         [1.0])


class TestAlignmentQuerierWithLSHNearNeighbor(BaseAlignmentQuerierTests,
        unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alignment_querier_subclass = alignment_query.AlignmentQuerierWithLSHNearNeighbor


class TestAlignmentQuerierWithKmerSharding(BaseAlignmentQuerierTests,
        unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alignment_querier_subclass = alignment_query.AlignmentQuerierWithKmerSharding




"""Tests for seq_io module.
"""

import logging
import tempfile
from os import unlink
import unittest

from adapt.utils import year_cover

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TestReadYears(unittest.TestCase):
    """Tests reading years from a file.
    """

    def setUp(self):
        # Disable logging
        logging.disable(logging.INFO)

    def test_read_success(self):
        # Write the temporary years file
        fn = tempfile.NamedTemporaryFile(mode='w', delete=False)
        fn.write("genome_1\t2016\n")
        fn.write("genome_2\t2016\n")
        fn.write("genome_3\t2018\n")
        fn.write("genome_4\t2016\n")
        fn.write("\n")
        fn.write("\n")
        # Closes the file so that it can be reopened on Windows
        fn.close()

        expected = {2016: {"genome_1", "genome_2", "genome_4"},
                    2018: {"genome_3"}}

        years = year_cover.read_years(fn.name)
        self.assertEqual(years, expected)        

        # Delete temporary file
        unlink(fn.name)

    def test_read_fail_nonyear(self):
        fn = tempfile.NamedTemporaryFile(mode='w', delete=False)
        fn.write("genome_1\t2016\n")
        fn.write("genome_2\t16\n")
        fn.write("genome_3\t2018\n")
        fn.write("genome_4\t2016\n")
        fn.write("\n")
        fn.write("\n")
        # Closes the file so that it can be reopened on Windows
        fn.close()
        with self.assertRaises(ValueError):
            years = year_cover.read_years(fn.name)
        # Delete temporary file
        unlink(fn.name)

    def test_read_fail_duplicate_sequence(self):
        fn = tempfile.NamedTemporaryFile(mode='w', delete=False)
        fn.write("genome_1\t2016\n")
        fn.write("genome_2\t2016\n")
        fn.write("genome_3\t2018\n")
        fn.write("genome_1\t2018\n")
        fn.write("\n")
        fn.write("\n")
        # Closes the file so that it can be reopened on Windows
        fn.close()
        with self.assertRaises(ValueError):
            years = year_cover.read_years(fn.name)
        # Delete temporary file
        unlink(fn.name)

    def tearDown(self):
        # Re-enable logging
        logging.disable(logging.NOTSET)


class TestConstructPartialCovers(unittest.TestCase):
    """Tests constructing partial cover for each year.
    """

    def setUp(self):
        # Disable logging
        logging.disable(logging.INFO)

    def test_construct_partial_covers_full_outside_range(self):
        cover_frac = year_cover.construct_partial_covers(
            {2013, 2014, 2015}, 2018, 1.0, 0.9)
        self.assertEqual(cover_frac,
            {2013: 0.9**5, 2014: 0.9**4, 2015:0.9**3})

    def test_construct_partial_covers_full_in_range(self):
        cover_frac = year_cover.construct_partial_covers(
            {2013, 2014, 2015, 2016, 2017}, 2015, 1.0, 0.9)
        self.assertEqual(cover_frac,
            {2013: 0.9**2, 2014: 0.9, 2015: 1.0, 2016: 1.0, 2017: 1.0})

    def test_construct_partial_covers_skip_years(self):
        cover_frac = year_cover.construct_partial_covers(
            {2013, 2014, 2017, 2018}, 2018, 0.5, 0.9)
        self.assertEqual(cover_frac,
            {2013: 0.5 * 0.9**5, 2014: 0.5 * 0.9**4, 2017: 0.5 * 0.9, 2018: 0.5})

    def tearDown(self):
        # Re-enable logging
        logging.disable(logging.NOTSET)




#!/usr/bin/env python3
"""Design guides naively for diagnostics.

This can be used as a baseline for the output of design.py.
"""

import argparse
import logging
import heapq

from adapt import alignment
from adapt.utils import log
from adapt.utils import seq_io

__author__ = 'Hayden Metsky <hmetsky@broadinstitute.org>, Priya P. Pillai <ppillai@broadinstitute.org>'

logger = logging.getLogger(__name__)


def construct_guide_naively_at_each_pos(aln, args, ref_seq=None):
    """Naively construct a guide sequence at each position of an alignment.

    This constructs a guide sequence at each position. It does so in two
    ways: 'consensus' (the consensus sequence at the position) and
    'mode' (the most common sequence at the position).

    Args:
        aln: alignment.Alignment object
        args: arguments to program
        ref_seq: reference sequence to base diversity guides on

    Returns:
        list x where x[i] is a dict, giving a guide sequence at position
        i of the alignment. x[i][method] is a tuple (guide, frac),
        where method is 'consensus' or 'mode', and guide gives a guide
        sequence (a string) constructed using the method and frac is
        the fraction of sequences in the alignment to which the guide
        binds
    """
    start_positions = range(aln.seq_length - args.guide_length + 1)
    guides = [None for _ in start_positions]
    ref_seq_aln = alignment.Alignment.from_list_of_seqs([ref_seq])
    for i in start_positions:
        # Extract the portion of the alignment that starts at i
        pos_start, pos_end = i, i + args.guide_length
        aln_for_guide = aln.extract_range(pos_start, pos_end)

        # When constructing guides, ignore any sequences in the alignment
        # that have a gap in this region. 
        seqs_with_gap = set(aln_for_guide.seqs_with_gap())
        seqs_to_consider = set(range(aln.num_sequences)) - seqs_with_gap
        # Only look at sequences with valid flanking regions
        seqs_to_consider = aln.seqs_with_required_flanking(i, args.guide_length,
                args.required_flanking_seqs, seqs_to_consider=seqs_to_consider)
        ref_seqs_to_consider = ref_seq_aln.seqs_with_required_flanking(
                i, args.guide_length, args.required_flanking_seqs)

        consensus_guide = None
        mode_guide = None
        diversity_guide = None

        frac_with_gap = float(len(seqs_with_gap)) / aln.num_sequences
        # Do not bother designing a guide here; there are
        # too many sequences with a gap
        if frac_with_gap < args.skip_gaps:
            # Construct guides
            # Only design guides if there exist sequences to consider
            if len(seqs_to_consider) > 0:
                if args.consensus:
                    consensus_guide = aln_for_guide.determine_consensus_sequence(
                            seqs_to_consider=seqs_to_consider)
                if args.mode:
                    mode_guide = aln_for_guide.determine_most_common_sequence(
                            seqs_to_consider=seqs_to_consider, skip_ambiguity=True) \

            if len(ref_seqs_to_consider) > 0:
                if args.diversity:
                    diversity_guide = ref_seq[pos_start:pos_end]

        # Determine the fraction of the sequences that each guide binds to
        if consensus_guide is not None:
            consensus_guide_bound = aln.sequences_bound_by_guide(
                    consensus_guide, i, args.guide_mismatches,
                    args.allow_gu_pairs, required_flanking_seqs=args.required_flanking_seqs)
            consensus_guide_frac = float(len(consensus_guide_bound)) / aln.num_sequences
        else:
            consensus_guide = 'None'
            consensus_guide_frac = 0

        if mode_guide is not None:
            mode_guide_bound = aln.sequences_bound_by_guide(
                    mode_guide, i, args.guide_mismatches,
                    args.allow_gu_pairs, required_flanking_seqs=args.required_flanking_seqs)
            mode_guide_frac = float(len(mode_guide_bound)) / aln.num_sequences
        else:
            mode_guide = 'None'
            mode_guide_frac = 0

        if diversity_guide is not None:
            diversity_guide_bound = aln.sequences_bound_by_guide(
                diversity_guide, i, args.guide_mismatches,
                args.allow_gu_pairs, required_flanking_seqs=args.required_flanking_seqs)
            diversity_guide_frac = float(len(diversity_guide_bound)) / aln.num_sequences
            if args.diversity == 'entropy':
                all_entropy = aln_for_guide.position_entropy()
                diversity_metric = sum(all_entropy)/args.guide_length
            else:
                raise ValueError("Invalid diversity method '%s'; use one of ['entropy']" %args.diversity)
        else:
            diversity_guide = 'None'
            diversity_guide_frac = 0
            diversity_metric = float('inf')

        d = {}
        if args.consensus:
            d['consensus'] = (consensus_guide, consensus_guide_frac)
        if args.mode:
            d['mode'] = (mode_guide, mode_guide_frac)
        if args.diversity:
            d[args.diversity] = ((diversity_guide, diversity_guide_frac), diversity_metric)
        guides[i] = d
    return guides


def find_guide_in_each_window(guides, aln_length, args, obj_type='max'):
    """Determine a guide for each window of an alignment.

    For each window, this selects the args.best_n guides within it 
    (given one guide per position) that has the best metric score.

    To break ties, this selects the first guide (in terms of
    position) among ones with a tied metric.

    Args:
        guides: list such that guides[i] is a tuple (guide, metric)
            giving a guide sequence (guide) at position i of
            an alignment and a metric (metric) that can be used to 
            compare guide quality
        aln_length: length of alignment
        args: arguments to program
        obj_type: if 'max', consider the largest value the best; else
            if 'min', consider the smallest value the best. Must be
            either 'min' or 'max'

    Returns:
        list x where x[i] gives a list of the args.best_n tuples (guide, metric)
        representing guides in the window that starts at position i
    """
    window_start_positions = range(aln_length - args.window_size + 1)
    guide_in_window = [[] for _ in window_start_positions]
    for i in window_start_positions:
        window_start, window_end = i, i + args.window_size
        last_guide_pos = window_end - args.guide_length
        logger.info("Searching for a guide within window [%d, %d)" %
                (window_start, window_end))

        # Check if any of the guides are no longer within the window
        # Keep track of which positions still have guides in the heap
        positions = set()
        if i > 0:
            for guide in guide_in_window[i-1]:
                if guide[1] >= window_start:
                    guide_in_window[i].append(guide)
                    positions.add(guide[1])
        heapq.heapify(guide_in_window[i])

        if len(guide_in_window[i]) < args.best_n: 
            # One of the previous args.best_n best guides is no longer in the 
            # window; find a new one
            for j in range(window_start, last_guide_pos + 1):
                # Skip if guide is already in the heap
                if -j in positions:
                    continue

                guide, metric = guides[j]
                # Reverse order for minimizing
                if obj_type == 'min':
                    metric = -metric

                # If there aren't args.best_n guides on the heap yet, add the 
                # guide to the heap
                if len(guide_in_window[i]) < args.best_n:
                    # Use negative position to break ties
                    heapq.heappush(guide_in_window[i], (metric, -j, guide))
                # If the new guide has a better metric than the worst guide on the
                # heap, remove the worst guide and add the new one
                elif metric > guide_in_window[i][0][0]:
                    # Use negative position to break ties
                    heapq.heappushpop(guide_in_window[i], (metric, -j, guide))
        else:
            # All args.best_n guides are still within the window, but now
            # check if the new guide at the very end of the window does better
            guide, metric = guides[last_guide_pos]
            # Reverse order for minimizing
            if obj_type == 'min':
                metric = -metric
            if metric > guide_in_window[i][0][0]:
                heapq.heappushpop(guide_in_window[i], (metric, -last_guide_pos, guide))

    # Undo reverse order for minimizing and sort
    fix = 1 if obj_type == 'max' else -1
    guide_in_window = [[(guide, fix*metric) for metric, _, guide \
            in sorted(guide_in_window_i, reverse=True)] \
            for guide_in_window_i in guide_in_window]
    return guide_in_window


def main(args):
    # Allow G-U base pairing, unless it is explicitly disallowed
    args.allow_gu_pairs = not args.do_not_allow_gu_pairing
    # Run the consensus method, unless it is explicitly disallowed
    args.consensus = not args.no_consensus
    # Run the mode method, unless it is explicitly disallowed
    args.mode = not args.no_mode
    # Check if there is a reference sequence if there is a diversity method
    if args.diversity:
        if not args.ref_seq:
            raise Exception('Must include a reference sequence label to run any diversity method')

    args.required_flanking_seqs = (args.require_flanking5, args.require_flanking3)

    # Read the input alignment
    seqs = seq_io.read_fasta(args.in_fasta)
    aln = alignment.Alignment.from_list_of_seqs(list(seqs.values()))
    try:
        ref_seq = seqs[args.ref_seq] if args.ref_seq else None
    except KeyError:
        raise Exception('Reference sequence %s does not match any label of the sequences in the given FASTA' \
                %args.ref_seq)

    # Construct a guide at each position of the alignment
    logger.info("Constructing guides naively at each position of alignment")
    guides = construct_guide_naively_at_each_pos(aln, args, ref_seq=ref_seq)

    # Find the best guide in each window (for the
    # consensus, mode, and diversity approaches)
    if args.consensus:
        logger.info("Searching for consensus guides")
        consensus_guides_in_window = find_guide_in_each_window(
                [guides[i]['consensus'] for i in range(len(guides))],
                aln.seq_length, args)
    if args.mode:
        logger.info("Searching for mode guides")
        mode_guides_in_window = find_guide_in_each_window(
                [guides[i]['mode'] for i in range(len(guides))],
                aln.seq_length, args)
    if args.diversity:
        logger.info("Searching for %s guides" %args.diversity)
        diversity_guides_in_window = find_guide_in_each_window(
                [guides[i][args.diversity] for i in range(len(guides))],
                aln.seq_length, args, obj_type='min')

    # Write the guides to a TSV file
    with open(args.out_tsv, 'w') as outf:
        header = ['window-start', 'window-end', 'rank']
        if args.consensus:
            header.extend(['target-sequence-by-consensus', 'frac-bound-by-consensus'])
        if args.mode:
            header.extend(['target-sequence-by-mode', 'frac-bound-by-mode'])
        if args.diversity:
            header.extend(['target-sequence-by-%s' %args.diversity, args.diversity, 'frac-bound-by-%s' %args.diversity])

        outf.write('\t'.join(header) + '\n')
        for i in range(aln.seq_length - args.window_size + 1):
            for j in range(args.best_n):
                line = [i, i + args.window_size, j+1]
                if args.consensus:
                    line.extend(consensus_guides_in_window[i][j])
                if args.mode:
                    line.extend(mode_guides_in_window[i][j])
                if args.diversity:
                    diversity_line = (diversity_guides_in_window[i][j][0][0], 
                                      diversity_guides_in_window[i][j][1],
                                      diversity_guides_in_window[i][j][0][1])
                    line.extend(diversity_line)
                outf.write('\t'.join([str(x) for x in line]) + '\n') 


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    # Input alignment and output file
    parser.add_argument('in_fasta',
            help=("Path to input FASTA (aligned)"))
    parser.add_argument('out_tsv',
            help=("Path to TSV file to which to write the output"))

    # Window size
    parser.add_argument('-w', '--window-size', type=int, default=200,
            help=("Output guide(s) within each window (sliding along "
                  "the alignment) of this length"))

    # Parameters on guide length and mismatches
    parser.add_argument('-gl', '--guide-length', type=int, default=28,
            help="Length of guide to construct")
    parser.add_argument('-gm', '--guide-mismatches', type=int, default=0,
            help=("Allow for this number of mismatches when "
                  "determining whether a guide covers a sequence"))

    # Best n guides per window
    parser.add_argument('--best-n', type=int, default=1,
            help=("Find the best BEST_N guides in each window"))

    # G-U pairing options
    parser.add_argument('--do-not-allow-gu-pairing', action='store_true',
            help=("When determining whether a guide binds to a region of "
                  "target sequence, do not count G-U (wobble) base pairs as "
                  "matching. Default is to tolerate G-U pairing: namely, "
                  "A in an output guide sequence matches G in the "
                  "target and C in an output guide sequence matches T "
                  "in the target (since the synthesized guide is the reverse "
                  "complement of the output guide sequence)"))

    # Options to skip
    parser.add_argument('--skip-gaps', type=float, default=0.5,
            help=("If this fraction or more of sequences at a position contain "
                  "a gap character, do not design a guide there"))

    # Reference sequence
    parser.add_argument('--ref-seq', type=str, default=None,
            help=("The label used in the FASTA file of the reference sequence "
                  "to design guides based on sequence diversity; required "
                  "for diversity method"))

    # Guide sequence methods
    parser.add_argument('--no-consensus', action='store_true',
            help=("If set, do not use the consensus method to determine guides; "
                  "otherwise, will use the consensus method"))
    parser.add_argument('--no-mode', action='store_true',
            help=("If set, do not use the mode method to determine guides; "
                  "otherwise, will use the mode method"))
    parser.add_argument('--diversity', type=str, default=None, choices=["entropy"],
            help=("A string of which diversity method to use to determine guides "
                  "('entropy'); None (default) to not use a diversity method. "
                  "'entropy' will calculate the average per position entropy of "
                  "each potential guide, then return the guides at the positions "
                  "with the lowest entropy; nucleotides are determined by the "
                  "reference sequence"))

    # Requiring flanking sequence (PFS)
    parser.add_argument('--require-flanking5',
        help=("Require the given sequence on the 5' protospacer flanking "
              "site (PFS) of each designed guide; this tolerates ambiguity "
              "in the sequence (e.g., 'H' requires 'A', 'C', or 'T', or, "
              "equivalently, avoids guides flanked by 'G'). Note that "
              "this is the 5' end in the target sequence (not the spacer "
              "sequence)."))
    parser.add_argument('--require-flanking3',
        help=("Require the given sequence on the 3' protospacer flanking "
              "site (PFS) of each designed guide; this tolerates ambiguity "
              "in the sequence (e.g., 'H' requires 'A', 'C', or 'T', or, "
              "equivalently, avoids guides flanked by 'G'). Note that "
              "this is the 3' end in the target sequence (not the spacer "
              "sequence)."))

    # Log levels
    parser.add_argument("--debug",
        dest="log_level",
        action="store_const",
        const=logging.DEBUG,
        default=logging.WARNING,
        help=("Debug output"))
    parser.add_argument("--verbose",
        dest="log_level",
        action="store_const",
        const=logging.INFO,
        help=("Verbose output"))

    args = parser.parse_args()

    log.configure_logging(args.log_level)
    main(args)



"""Tests for primer_search module.
"""

import logging
import random
import unittest

from adapt import alignment
from adapt import primer_search
from adapt.utils import guide

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TestPrimerResult(unittest.TestCase):
    """Tests methods in the PrimerResult class.
    """

    def test_does_overlap(self):
        a = primer_search.PrimerResult(10, 1, 5, 1.0, {'AAAAA'})
        self.assertTrue(a.overlaps(a))

        b = primer_search.PrimerResult(12, 1, 5, 1.0, {'AAAAA'})
        self.assertTrue(a.overlaps(b))
        self.assertTrue(b.overlaps(a))

    def test_does_not_overlap(self):
        a = primer_search.PrimerResult(10, 1, 5, 1.0, {'AAAAA'})
        b = primer_search.PrimerResult(15, 1, 5, 1.0, {'AAAAA'})
        c = primer_search.PrimerResult(16, 1, 5, 1.0, {'AAAAA'})
        self.assertFalse(a.overlaps(b))
        self.assertFalse(b.overlaps(a))
        self.assertFalse(a.overlaps(c))
        self.assertFalse(c.overlaps(a))

    def test_does_overlap_with_expand(self):
        a = primer_search.PrimerResult(10, 1, 5, 1.0, {'AAAAA'})
        self.assertTrue(a.overlaps(a, expand=5))

        b = primer_search.PrimerResult(12, 1, 5, 1.0, {'AAAAA'})
        self.assertTrue(a.overlaps(b, expand=5))
        self.assertTrue(b.overlaps(a, expand=5))

        b = primer_search.PrimerResult(17, 1, 5, 1.0, {'AAAAA'})
        self.assertTrue(a.overlaps(b, expand=5))
        self.assertTrue(b.overlaps(a, expand=5))

        c = primer_search.PrimerResult(2, 1, 5, 1.0, {'AAAAA'})
        self.assertTrue(a.overlaps(c, expand=5))
        self.assertTrue(c.overlaps(a, expand=5))

    def test_does_not_overlap_with_expand(self):
        a = primer_search.PrimerResult(10, 1, 5, 1.0, {'AAAAA'})
        b = primer_search.PrimerResult(20, 1, 5, 1.0, {'AAAAA'})
        c = primer_search.PrimerResult(0, 1, 5, 1.0, {'AAAAA'})
        self.assertFalse(a.overlaps(b, expand=5))
        self.assertFalse(b.overlaps(a, expand=5))
        self.assertFalse(a.overlaps(c, expand=5))
        self.assertFalse(c.overlaps(a, expand=5))

    def test_does_overlap_range(self):
        a = primer_search.PrimerResult(10, 1, 5, 1.0, {'AAAAA'})
        self.assertTrue(a.overlaps_range(5, 20))
        self.assertTrue(a.overlaps_range(5, 12))
        self.assertTrue(a.overlaps_range(12, 20))
        self.assertTrue(a.overlaps_range(11, 13))

    def test_does_not_overlap_range(self):
        a = primer_search.PrimerResult(10, 1, 5, 1.0, {'AAAAA'})
        self.assertFalse(a.overlaps_range(2, 8))
        self.assertFalse(a.overlaps_range(20, 25))


class TestPrimerSearch(unittest.TestCase):
    """Tests methods in the PrimerSearch class.
    """

    def setUp(self):
        # Disable logging
        logging.disable(logging.WARNING)

        # Set a random seed so hash functions are always the same
        random.seed(0)

        self.a_seqs = ['ATCGAA',
                       'ATCGAT',
                       'TTCGAA',
                       'CTCGAT',
                       'GGCGAA']
        self.a_aln = alignment.Alignment.from_list_of_seqs(self.a_seqs)
        self.a = primer_search.PrimerSearcher(self.a_aln, 4, 0, 1.0, (1, 1, 100))
        self.a_gc_bounds = primer_search.PrimerSearcher(self.a_aln,
                4, 0, 1.0, (1, 1, 100), primer_gc_content_bounds=(0.4, 0.6))

    def _pr(self, start, num_primers, frac_bound, primers_in_cover):
        # Construct a PrimerResult object
        return primer_search.PrimerResult(start, num_primers, 4, frac_bound,
            primers_in_cover)

    def test_find_primers_simple(self):
        covers = list(self.a.find_primers())
        expected = [self._pr(0, 4, 1.0, set(['ATCG', 'TTCG', 'CTCG', 'GGCG'])),
                    self._pr(1, 2, 1.0, set(['TCGA', 'GCGA'])),
                    self._pr(2, 2, 1.0, set(['CGAA', 'CGAT']))]
        self.assertEqual(covers, expected)

    def test_find_primers_with_max(self):
        covers = list(self.a.find_primers(max_at_site=2))
        expected = [self._pr(1, 2, 1.0, set(['TCGA', 'GCGA'])),
                    self._pr(2, 2, 1.0, set(['CGAA', 'CGAT']))]
        self.assertEqual(covers, expected)

    def test_find_primers_with_gc_bounds(self):
        covers = list(self.a_gc_bounds.find_primers())
        expected = [self._pr(2, 2, 1.0, set(['CGAA', 'CGAT']))]
        self.assertEqual(covers, expected)

    def test_find_primers_with_grouped_cover_frac(self):
        cover_frac = {0: 1.0, 1: 0.01, 2: 1.0}
        seq_groups = {0: {0, 1}, 1: {2, 3}, 2: {4}}
        self.a1 = primer_search.PrimerSearcher(self.a_aln, 4, 0, cover_frac,
                (1, 1, 100),
                seq_groups=seq_groups)
        covers = list(self.a1.find_primers())
        # Two possibilities for group 1 (first primer is 'TTCG' or 'CTCG')
        expected1 = [self._pr(0, 3, 4.0/5.0, set(['ATCG', 'GGCG', 'TTCG'])),
                     self._pr(1, 2, 5.0/5.0, set(['TCGA', 'GCGA'])),
                     self._pr(2, 2, 5.0/5.0, set(['CGAA', 'CGAT']))]
        expected2 = [self._pr(0, 3, 4.0/5.0, set(['ATCG', 'GGCG', 'CTCG'])),
                     self._pr(1, 2, 5.0/5.0, set(['TCGA', 'GCGA'])),
                     self._pr(2, 2, 5.0/5.0, set(['CGAA', 'CGAT']))]
        self.assertIn(covers, [expected1, expected2])

    def test_check_gc_content(self):
        self.assertTrue(self.a_gc_bounds.check_gc_content({'ATCG', 'GGTT'}))
        self.assertTrue(self.a_gc_bounds.check_gc_content({'ATCG', 'AGCT'}))
        self.assertFalse(self.a_gc_bounds.check_gc_content({'ATCA', 'AGCT'}))
        self.assertFalse(self.a_gc_bounds.check_gc_content({'ATCG', 'AGAT'}))
        self.assertFalse(self.a_gc_bounds.check_gc_content({'ATCA', 'AGAT'}))
        self.assertFalse(self.a_gc_bounds.check_gc_content({'GGCA', 'AGCT'}))

    def tearDown(self):
        # Re-enable logging
        logging.disable(logging.NOTSET)
